from mgz.fast.header import decompress, parse_version
from mgz.summary.full import FullSummary
from mgz.model.compat import ModelSummary
from mgz.util import Version
import logging
import zlib

logger = logging.getLogger(__name__)

class SummaryStub:

    def __call__(self, data, fallback=False):
        try:
            header = decompress(data)
            version, game, save, log = parse_version(header, data)
            data.seek(0)
            supported = (version is Version.DE and save > 13.34) # or version is Version.USERPATCH15
        except zlib.error:
            supported = False
        if supported and not fallback:
            logger.info("using model summary")
            try:
                return ModelSummary(data)
            except RuntimeError as e:
                logger.warning(f"could not fast parse; falling back: {e}")
        logger.info("using full summary")
        return FullSummary(data)


Summary = SummaryStub()



"""Replay."""

from construct import Array, Byte, Flag, Float32l, Int16ul, Int32sl, Int32ul, Padding, Struct, If, Embedded

from mgz.util import Version, find_save_version

# pylint: disable=invalid-name


# Basic information about the recorded game
replay = "replay"/Struct(
    "old_time"/Int32ul,
    "world_time"/Int32ul,
    "old_world_time"/Int32ul,
    "game_speed_id"/Int32ul, # world_time_delta
    "world_time_delta_seconds"/Int32ul,
    "timer"/Float32l,
    "game_speed_float"/Float32l,
    "temp_pause"/Byte,
    "next_object_id"/Int32ul,
    "next_reusable_object_id"/Int32sl,
    "random_seed"/Int32ul,
    "random_seed_2"/Int32ul,
    "rec_player"/Int16ul, # id of the rec owner
    "num_players"/Byte, # including gaia
    Embedded(If(lambda ctx: ctx._.version != Version.AOK, Struct(
        "instant_build"/Flag,
        "cheats_enabled"/Flag
    ))),
    "game_mode"/Int16ul, # MP or SP?
    "campaign"/Int32ul,
    "campaign_player"/Int32ul,
    "campaign_scenario"/Int32ul,
    "king_campaign"/Int32ul,
    "king_campaign_player"/Byte,
    "king_campaign_scenario"/Byte,
    "player_turn"/Int32ul,
    "player_time_delta"/Array(lambda obj: obj.num_players if find_save_version(obj) >= 61.5 else 9, "turn"/Int32ul),
    If(lambda ctx: ctx._.version == Version.DE, Padding(8))
)



"""Determine dataset."""

import mgz
from mgz.util import Version
from mgz.reference import get_dataset


def resolve_hd_version(hd, save_version):
    """Best guess at HD version."""
    if hd.version == 1006:
        if 'test_57' in hd and hd.test_57.is_57:
            return '5.7'
        else:
            return '5.8'
    if hd.version == 1005:
        return '>=5.0,<5.7'
    if hd.version == 1004:
        return '4.8'
    if save_version >= 12.36:
        return '>=4.6,<4.8'
    return None


def get_dataset_data(header):
    """Get dataset."""
    sample = header.initial.players[0].attributes.player_stats
    mod = None
    if header.de:
        mod = header.de.dlc_ids
    if 'mod' in sample:
        mod = (sample.mod.get('id'), sample.mod.get('version'))
    _, ref = get_dataset(header.version, mod)
    if header.version == Version.DE:
        if 11 in header.de.dlc_ids:
            return {
                'id': 101,
                'name': 'Return of Rome',
                'version': None
            }, ref
        return {
            'id': 100,
            'name': 'Definitive Edition',
            'version': None
        }, ref
    elif header.version == Version.HD:
        return {
            'id': 300,
            'name': 'HD Edition',
            'version': resolve_hd_version(header.hd, header.save_version)
        }, ref
    if 'mod' in sample and sample.mod['id'] == 0 and sample.mod['version'] == '2':
        raise ValueError("invalid mod version")
    if 'mod' in sample and sample.mod['id'] > 0:
        return sample.mod, ref
    if 'trickle_food' in sample and sample.trickle_food:
        return {
            'id': 1,
            'name': mgz.const.MODS.get(1),
            'version': '<5.7.2'
        }, ref
    if header.version == Version.AOK:
        return {
            'id': 200,
            'name': 'Age of Kings',
            'version': '2.0a'
        }, ref
    if header.version == Version.AOC10:
        return {
            'id': 0,
            'name': 'The Conquerors',
            'version': '1.0'
        }, ref
    return {
        'id': 0,
        'name': 'The Conquerors',
        'version': '1.0c'
    }, ref



"""Unit Types."""
from construct import (
    Struct, Float32l, Int16ul, Byte, Embedded, Switch, If,
    Int32ul, Array, String, Padding, IfThenElse, this
)
from mgz.util import find_version, find_save_version, Version

attribute = "attribute"/Struct(
    "type"/Int16ul,
    "amount"/Float32l,
    "flag"/Byte
)

damage_sprite = "damage_sprite"/Struct(
    "sprite"/Int16ul,
    "damage_percent"/Int16ul,
    "flag"/Byte
)

base = "base"/Struct(
    "name_len"/Int16ul,
    "id"/Int16ul,
    "string_id"/Int16ul,
    "string_id_2"/Int16ul,
    "unit_class"/Int16ul,
    "standing_sprite_1"/Int16ul,
    "standing_sprite_2"/Int16ul,
    "dying_sprite"/Int16ul,
    "undead_sprite"/Int16ul,
    "undead_flag"/Byte,
    "hitpoints"/Int16ul,
    "los"/Float32l,
    "garrison_capacity"/Byte,
    "radius"/Struct(
        "x"/Float32l,
        "y"/Float32l,
        "z"/Float32l
    ),
    "train_sound"/Int16ul,
    "damage_sound"/Int16ul,
    "death_spawn"/Int16ul,
    "sort_number"/Byte,
    "can_be_built_on"/Byte,
    "button_picture"/Int16ul,
    "hide_in_editor"/Byte,
    "portrait"/Int16ul,
    "enabled"/Byte,
    "disabled"/Byte,
    "tile_req"/Struct(
        "x"/Int16ul,
        "y"/Int16ul
    ),
    "center_tile_req"/Struct(
        "x"/Int16ul,
        "y"/Int16ul
    ),
    "construction_radius"/Struct(
        "x"/Float32l,
        "y"/Float32l
    ),
    "elevation_flag"/Byte,
    "fog_flag"/Byte,
    "terrain_restriction_id"/Int16ul,
    "movement_type"/Byte,
    "attribute_max"/Int16ul,
    "attribute_rot"/Float32l,
    "area_effect_level"/Byte,
    "combat_level"/Byte,
    "select_level"/Byte,
    "map_draw_level"/Byte,
    "unit_level"/Byte,
    "multiple_attribute_mod"/Float32l,
    "map_color"/Byte,
    "help_string_id"/Int32ul,
    "help_page_id"/Int32ul,
    "hotkey_id"/Int32ul,
    "recyclable"/Byte,
    "track_as_resource"/Byte,
    "create_doppleganger"/Byte,
    "resource_group"/Byte,
    "occlusion_mask"/Byte,
    "obstruction_type"/Byte,
    "selection_shape"/Byte,
    "object_flags"/If(lambda ctx: find_version(ctx) != Version.AOK, Int32ul),
    "civilization"/Byte,
    "attribute_piece"/Byte,
    "outline_radius"/Struct(
        "x"/Float32l,
        "y"/Float32l,
        "z"/Float32l
    ),
    "attributes"/Array(3, attribute),
    "num_damage_sprites"/Byte,
    "damage_sprites"/Array(this.num_damage_sprites, damage_sprite),
    "selected_sound"/Int16ul,
    "death_sound"/Int16ul,
    "attack_reaction"/Byte,
    "convert_terrain_flag"/Byte,
    "name"/String(this.name_len),
    "copy_id"/Int16ul,
    "group"/Int16ul
)

animated = "animated"/Struct(
    base,
    "speed"/Float32l
)

moving = "moving"/Struct(
    animated,
    "move_sprite"/Int16ul,
    "run_sprite"/Int16ul,
    "turn_speed"/Float32l,
    "size_class"/Byte,
    "trailing_unit"/Int16ul,
    "trailing_options"/Byte,
    "trailing_spacing"/Float32l,
    "move_algorithm"/Byte,
    "turn_radius"/Float32l,
    "turn_radius_speed"/Float32l,
    "maximum_yaw_per_second_moving"/Float32l,
    "stationary_yaw_per_revolution_time"/Float32l,
    "maximum_yaw_per_second_stationary"/Float32l
)

action = "action"/Struct(
    moving,
    "default_task"/Int16ul,
    "search_radius"/Float32l,
    "work_rate"/Float32l,
    "handicap_work_rate"/If(lambda ctx: find_save_version(ctx) >= 25.06, Float32l),
    "drop_site"/Int16ul,
    "backup_drop_site"/Int16ul,
    "task_by_group"/Byte,
    "command_sound"/Int16ul,
    "move_sound"/Int16ul,
    "run_pattern"/Byte
)

weapon = "weapon"/Struct(
    "type"/Int16ul,
    "value"/Int16ul
)

base_combat = "base_combat"/Struct(
    action,
    "base_armour"/IfThenElse(lambda ctx: find_version(ctx) != Version.AOK, Int16ul, Byte),
    "num_weapons"/Int16ul,
    "weapons"/Array(this.num_weapons, weapon),
    "num_armours"/Int16ul,
    "armours"/Array(this.num_armours, weapon),
    "defense_terrain_bonus"/Int16ul,
    "weapon_range_max"/Float32l,
    "area_effect_range"/Float32l,
    "attack_speed"/Float32l,
    "missile_id"/Int16ul,
    "base_hit_chance"/Int16ul,
    "break_off_combat"/Byte,
    "frame_delay"/Int16ul,
    "weapon_offset"/Struct(
        "x"/Float32l,
        "y"/Float32l,
        "z"/Float32l
    ),
    "blast_level_offense"/Byte,
    "weapon_range_min"/Float32l,
    "missed_missile_spread"/Float32l,
    "fight_sprite"/Int16ul,
    "displayed_armour"/Int16ul,
    "displayed_attack"/Int16ul,
    "displayed_range"/Float32l,
    "displayed_reload_time"/Float32l
)

missile = "missile"/Struct(
    base_combat,
    "missile_type"/Byte,
    "targeting_type"/Byte,
    "missile_hit_info"/Byte,
    "missile_die_info"/Byte,
    "area_effect_specials"/Byte,
    "ballistics_ratio"/Float32l
)

attribute_cost = "cost"/Struct(
    "type"/Int16ul,
    "amount"/Int16ul,
    "flag"/Byte,
    Padding(1)
)

combat = "combat"/Struct(
    base_combat,
    "cost"/Array(3, attribute_cost),
    "create_time"/Int16ul,
    "create_at_building"/Int16ul,
    "creat_button"/Byte,
    "rear_attack_modifieer"/Float32l,
    "flank_attack_modifier"/Float32l,
    "tribe_unit_type"/Byte,
    "hero_flag"/Byte,
    "garrison_sprite"/Int32ul,
    "volley_fire_aount"/Float32l,
    "max_attacks_in_volley"/Byte,
    "volley_spread"/Struct(
        "x"/Float32l,
        "y"/Float32l
    ),
    "volley_start_spread_adjustment"/Float32l,
    "volley_missile"/Int32ul,
    "special_attack_sprite"/Int32ul,
    "special_attack_flag"/Byte,
    "displayed_pierce_armour"/Int16ul
)

linked_building = "link"/Struct(
    "unit_id"/Int16ul,
    "x_offset"/Float32l,
    "y_offset"/Float32l
)

building = "building"/Struct(
    combat,
    "construction_sprite"/Int16ul,
    "snow_sprite"/If(lambda ctx: find_version(ctx) != Version.AOK, Int16ul),
    "connect_flag"/Byte,
    "facet"/Int16ul,
    "destroy_on_build"/Byte,
    "on_build_make_unit"/Int16ul,
    "on_build_make_tile"/Int16ul,
    "on_build_make_overlay"/Int16ul,
    "on_build_make_tech"/Int16ul,
    "can_burn"/Byte,
    "linked_buildings"/Array(4, linked_building),
    "construction_unit"/Int16ul,
    "transform_unit"/Int16ul,
    "transform_sound"/Int16ul,
    "construction_sound"/Int16ul,
    "garrison_type"/Byte,
    "garrison_heal_rate"/Float32l,
    "garrison_repair_rate"/Float32l,
    "salvage_unit"/Int16ul,
    "salvage_attributes"/Array(6, Byte)
)

unit_type = "unit_type"/Struct(
    "type"/Byte,
    Embedded("properties"/Switch(this.type, {
        10: base,
        15: base,
        20: animated,
        25: animated,
        30: moving,
        40: action,
        50: base_combat,
        60: missile,
        70: combat,
        80: building
    }))
)



"""AI."""

from mgz import Version
from mgz.util import Find
from construct import (Array, Byte, If, Int16ul, Int32sl, Int32ul, Padding,
                       PascalString, Struct, this, IfThenElse)

# pylint: disable=invalid-name

script = "script"/Struct(
    Padding(4),
    "seq"/Int32sl,
    "max_rules"/Int16ul,
    "num_rules"/Int16ul,
    Padding(4),
    Array(this.num_rules, "rules"/Struct(
        Padding(12),
        "num_facts"/Byte,
        "num_facts_actions"/Byte,
        Padding(2),
        Array(16, "data"/Struct(
            "type"/Int32ul,
            "id"/Int16ul,
            Padding(2),
            Array(4, "params"/Int32ul)
        ))
    ))
)

ai = "ai"/Struct(
    "has_ai"/Int32ul, # if true, parse AI
    "yep"/If(
        this.has_ai == 1,
        IfThenElse(
            lambda ctx: ctx._.version == Version.DE,
            Find(b'\00' * 4096, None),  # The ai structure in DE seems to have changed, for now we simply skip it
            "ais"/Struct(
                "max_strings"/Int16ul,
                "num_strings"/Int16ul,
                Padding(4),
                Array(this.num_strings, "strings"/PascalString(lengthfield="name_length"/Int32ul,
                                                               encoding='latin1')),
                Padding(6),
                Array(8, script),
                Padding(104),
                Array(80, "timers"/Int32sl),
                Array(256, "shared_goals"/Int32sl),
                Padding(4096),
            )
        )
    )
)



"""Map summary."""

import re
import mgz
from mgz.util import Version


ENCODING_MARKERS = [
    ['Map Type: ', 'latin-1', 'en'],
    ['Map type: ', 'latin-1', 'en'],
    ['Location: ', 'utf-8', 'en'],
    ['Tipo de mapa: ', 'latin-1', 'es'],
    ['Ubicación: ', 'utf-8', 'es'],
    ['Ubicaci: ', 'utf-8', 'es'],
    ['Local: ', 'utf-8', 'es'],
    ['Kartentyp: ', 'latin-1', 'de'],
    ['Karte: ', 'utf-8', 'de'],
    ['Art der Karte: ', 'latin-1', 'de'],
    ['Type de carte\xa0: ', 'latin-1', 'fr'],
    ['Emplacement :', 'utf-8', 'fr'],
    ['Type de carte : ', 'latin-1', 'fr'],
    ['Tipo di mappa: ', 'latin-1', 'it'],
    ['Posizione: ', 'utf-8', 'it'],
    ['Tipo de Mapa: ', 'latin-1', 'pt'],
    ['Kaarttype', 'latin-1', 'nl'],
    ['Lokalizacja: ', 'utf-8', 'pl'],
    ['Harita Türü: ', 'ISO-8859-1', 'tr'],
    ['Harita Sitili', 'ISO-8859-1', 'tr'],
    ['Harita tipi', 'ISO-8859-1', 'tr'],
    ['Konum: ', 'ISO-8859-1', 'tr'],
    ['??? ?????: ', 'ascii', 'tr'], # corrupt lang dll?
    ['Térkép tipusa', 'ISO-8859-1', 'hu'],
    ['Typ mapy: ', 'ISO-8859-2', None],
    ['Тип карты: ', 'windows-1251', 'ru'],
    ['Тип Карты: ', 'windows-1251', 'ru'],
    ['Расположение: ', 'utf-8', 'ru'],
    ['マップの種類: ', 'SHIFT_JIS', 'jp'],
    ['マップ ', 'utf-8', 'jp'],
    ['지도 종류: ', 'cp949', 'kr'],
    ['地??型', 'big5', 'zh'],
    ['地图类型: ', 'cp936', 'zh'],
    ['地圖類別：', 'cp936', 'zh'],
    ['地圖類別：', 'big5', 'zh'],
    ['地图类别：', 'cp936', 'zh'],
    ['地图类型：', 'GB2312', 'zh'],
    ['颌玉拙墁：', 'cp936', 'zh'],
    ['位置：', 'utf-8', 'zh'],
    ['舞台: ', 'utf-8', 'zh'],
    ['Vị trí: ', 'utf-8', 'vi'],
    ['위치: ', 'utf-8', 'kr'],
    ['Τύπος Χάρτη: ', 'utf-8', 'gr'],
    ['Emplacement: ', 'utf-8', 'fr'],
    ['Local: ', 'utf-8', 'pt'],
    ['Mapa: ', 'utf-8', 'cs'],
    ['Plats: ', 'utf-8', 'se']
]
LANGUAGE_MARKERS = [
    ['Dostepne', 'ISO-8859-2', 'pl'],
    ['oszukiwania', 'ISO-8859-2', 'pl'],
    ['Dozwoli', 'ISO-8859-2', 'pl'],
    ['Povol', 'ISO-8859-2', 'cs'], # Povolené, Povolit
    ['Mozno', 'ISO-8859-2', 'sk'],
    ['Dobývací', 'ISO-8859-2', 'cs']
]
WATER_TERRAIN = {
    0: [1, 4, 15, 22, 23],
    1: [1, 4, 11, 15, 22, 23],
    7: [1, 4, 15, 22, 23],
    100: [1, 4, 15, 22, 23, 26, 54, 57, 58, 59, 93, 94, 95, 96, 97, 98, 99]
}


def extract_from_instructions(instructions):
    """Extract data from instructions."""
    language = None
    encoding = 'unknown'
    name = 'Unknown'
    for pair in ENCODING_MARKERS:
        marker = pair[0]
        test_encoding = pair[1]
        e_m = marker.encode(test_encoding)
        for line in instructions.split(b'\n'):
            has = line.startswith(e_m)
            if has:
                pos = 0
                encoding = test_encoding
                name = line[pos+len(e_m):].decode(encoding).replace('.rms', '')
                language = pair[2]
                break

    # disambiguate certain languages
    if not language:
        language = 'unknown'
        for pair in LANGUAGE_MARKERS:
            if instructions.find(pair[0].encode(pair[1])) > -1:
                language = pair[2]
                break
    if encoding == 'unknown':
        raise ValueError('could not detect encoding')
    return encoding, language, name


def lookup_name(map_id, name, version, reference):
    """Lookup base game map if applicable."""
    custom = True
    is_de = version == Version.DE
    is_hd = version == Version.HD
    # 59: RM custom
    # 137: custom map pool
    # 138: DM custom
    if (map_id != 44 and not (is_de or is_hd)) or (map_id not in [59, 137, 138] and (is_de or is_hd)):
        map_keys = [int(k) for k in reference['maps'].keys()]
        if map_id in map_keys:
            name = reference['maps'][str(map_id)]
        elif version == Version.AOK:
            return name, False
        else:
            raise ValueError('unspecified builtin map: {} aka {} ({})'.format(map_id, name, reference['dataset']['name']))
        custom = False
    return name, custom


def get_map_seed(instructions):
    """Extract map seed from instructions."""
    match = re.search(rb'\x00.*? (\-?[0-9]+)\x00.*?\.rms', instructions)
    seed = None
    if match:
        seed = int(match.group(1))
    return seed


def get_modes(name):
    """Extract userpatch modes."""
    has_modes = name.find(': !')
    mode_string = ''
    if has_modes > -1:
        mode_string = name[has_modes + 3:]
        name = name[:has_modes]
    modes = {
        'direct_placement': 'P' in mode_string,
        'effect_quantity': 'C' in mode_string,
        'guard_state': 'G' in mode_string,
        'fixed_positions': 'F' in mode_string
    }
    return name, modes


def get_tiles(tiles, dimension):
    """Get map tile data."""
    tile_x = 0
    tile_y = 0
    for tile in tiles:
        if tile_x == dimension:
            tile_x = 0
            tile_y += 1
        yield {
            'x': tile_x,
            'y': tile_y,
            'terrain_id': tile[0],
            'elevation': tile[1]
        }
        tile_x += 1


def get_water_percent(tiles, dataset_id):
    """Get percent of map that is passable by ships."""
    if dataset_id not in WATER_TERRAIN:
        return None
    count = 0
    for tile in tiles:
        if tile[0] in WATER_TERRAIN[dataset_id]:
            count +=1
    return count/len(tiles)


def get_mod_id(strings):
    """Extract RMS mod id."""
    for obj in strings:
        if not obj.string:
            continue
        s = obj.string.value.decode('utf-8').split(':')
        if s[0] == 'SUBSCRIBEDMODS' and s[1] == 'RANDOM_MAPS':
            return s[3].split('_')[0]


def get_map_data(map_id, instructions, dimension, version, dataset_id, reference, tiles, de_seed=None, de_strings=[]):
    """Get the map metadata."""
    if instructions == b'\x00':
        raise ValueError('empty instructions')
    
    encoding, language, name = extract_from_instructions(instructions)
    name, custom = lookup_name(map_id, name, version, reference)
    seed = get_map_seed(instructions)
    name, modes = get_modes(name)

    return {
        'id': map_id if not custom else None,
        'name': name.strip(),
        'size': mgz.const.MAP_SIZES.get(dimension),
        'dimension': dimension,
        'seed': de_seed if de_seed else seed,
        'mod_id': get_mod_id(de_strings) if custom else None,
        'modes': modes,
        'custom': custom,
        'zr': name.startswith('ZR@'),
        'tiles': list(get_tiles(tiles, dimension)),
        'water': get_water_percent(tiles, dataset_id)
    }, encoding, language



"""Enumerations."""

from construct import Enum, Pass

# pylint: disable=invalid-name


def ObjectEnum(ctx):
    """Object Enumeration.

    Should export the whole list from the game for the best accuracy.
    """
    return Enum(
        ctx,
        villager_male=83,
        villager_female=293,
        scout_cavalry=448,
        eagle_warrior=751,
        king=434,
        flare=332,
        relic=285,
        turkey=833,
        sheep=594,
        deer=65,
        boar=48,
        iron_boar=810,
        ostrich=1026,
        javelina=822,
        crocodile=1031,
        rhinoceros=1139,
        wolf=126,
        jaguar=812,
        hawk=96,
        macaw=816,
        shore_fish=69,
        fish_1=455,
        fish_2=456,
        fish_4=458,
        fish_3=457,
        marlin_1=450,
        marlin_2=451,
        dolphin=452,
        cactus=709,
        berry_bush=59,
        stone_pile=102,
        gold_pile=66,
        forest_tree=350,
        forest_tree_2=411,
        snow_pine_tree=413,
        straggler_tree=349,
        tc_1=109,
        tc_2=618,
        tc_3=619,
        tc_4=620,
        castle=70,
        palisade_wall=72,
        stone_wall=117,
        stone_gate_1=64,
        stone_gate_2=81,
        stone_gate_3=88,
        stone_gate_4=95,
        palisade_gate_1=662,
        palisade_gate_2=666,
        palisade_gate_3=670,
        palisade_gate_4=674,
        fortified_wall=155,
        cliff_1=264,
        cliff_2=265,
        cliff_3=266,
        cliff_4=267,
        cliff_5=268,
        cliff_6=269,
        cliff_7=270,
        cliff_8=271,
        cliff_9=272,
        cliff_10=273,
        outpost=598,
        shipwreck=722,
        map_revealer=837,
        default=Pass
    )

def BuildingEnum(ctx):
    """Buildings Enumeration (in Action context)."""
    return Enum(
        ctx,
        dock=45,
        tc=621,
        default=Pass
    )

def GameTypeEnum(ctx):
    """Game Type Enumeration."""
    return Enum(
        ctx,
        RM=0,
        Regicide=1,
        DM=2,
        Scenario=3,
        Campaign=4,
        KingOfTheHill=5,
        WonderRace=6,
        DefendTheWonder=7,
        TurboRandom=8,
        CaptureTheRelic=10,
        SuddenDeath=11,
        BattleRoyale=12,
        EmpireWars=13,
        CoOpCampaign=15
    )

def ObjectTypeEnum(ctx):
    """Object Type Enumeration."""
    return Enum(
        ctx,
        static=10,
        animated=20,
        doppelganger=25,
        moving=30,
        action=40,
        base=50,
        missile=60,
        combat=70,
        building=80,
        tree=90,
        default=Pass
    )

def PlayerTypeEnum(ctx):
    """Player Type Enumeration."""
    return Enum(
        ctx,
        absent=0,
        closed=1,
        human=2,
        eliminated=3,
        computer=4,
        cyborg=5,
        spectator=6
    )

def DifficultyEnum(ctx):
    """Difficulty Enumeration."""
    return Enum(
        ctx,
        hardest=0,
        hard=1,
        moderate=2,
        standard=3,
        easiest=4,
        extreme=5,
        unknown=6
    )

def OperationEnum(ctx):
    """Operation Enumeration."""
    return Enum(
        ctx,
        action=1,
        sync=2,
        viewlock=3,
        chat=4,
        default="embedded"
    )

def ResourceEnum(ctx):
    """Resource Type Enumeration."""
    return Enum(
        ctx,
        food=0,
        wood=1,
        stone=2,
        gold=3,
        decay=12,
        fish=17,
        default=Pass # lots of resource types exist
    )

def VictoryEnum(ctx):
    """Victory Type Enumeration."""
    return Enum(
        ctx,
        standard=0,
        conquest=1,
        exploration=2,
        ruins=3,
        artifacts=4,
        discoveries=5,
        gold=6,
        time_limit=7,
        score=8,
        standard2=9,
        regicide=10,
        last_man=11
    )

def ResourceLevelEnum(ctx):
    """Resource Level Enumeration."""
    return Enum(
        ctx,
        none=-1,
        standard=0,
        low=1,
        medium=2,
        high=3,
        unknown1=4,
        unknown2=5,
        unknown3=6
    )

def RevealMapEnum(ctx):
    """Reveal Map Enumeration."""
    return Enum(
        ctx,
        normal=0,
        explored=1,
        all_visible=2,
        no_fog=3,
    )

def AgeEnum(ctx):
    """Age Enumeration."""
    return Enum(
        ctx,
        what=-2,
        unset=-1,
        dark=0,
        feudal=1,
        castle=2,
        imperial=3,
        postimperial=4,
        dmpostimperial=6,
        default='unknown'
    )

def TheirDiplomacyEnum(ctx):
    """Other Player's Diplomacy Enumeration."""
    return Enum(
        ctx,
        ally_or_self=0,
        enemy=3,
        treaty=4
    )

def DiplomacyStanceEnum(ctx):
    """Diplomacy stance."""
    return Enum(
        ctx,
        allied=0,
        neutral=1,
        enemy=3
    )

def GameActionModeEnum(ctx):
    """Game Action Modes."""
    return Enum(
        ctx,
        diplomacy=0,
        speed=1,
        instant_build=2,
        quick_build=4,
        allied_victory=5,
        cheat=6,
        unk0=9,
        spy=10,
        unk1=11,
        farm_queue=13,
        farm_unqueue=14,
        farm_autoqueue=16,
        fishtrap_queue=17,
        fishtrap_unqueue=18,
        fishtrap_autoqueue=19,
        default_stance=20,
        default=Pass
    )

def OrderTypeEnum(ctx):
    """Types of Orders."""
    return Enum(
        ctx,
        packtreb=1,
        unpacktreb=2,
        garrison=5,
        default=Pass
    )

def ReleaseTypeEnum(ctx):
    """Types of Releases."""
    return Enum(
        ctx,
        all=0,
        selected=3,
        sametype=4,
        notselected=5,
        inversetype=6,
        default=Pass
    )

def StanceEnum(ctx):
    """Types of stances."""
    return Enum(
        ctx,
        aggressive=0,
        defensive=1,
        stand_ground=2,
        passive=3
    )

def FormationEnum(ctx):
    """Types of formations."""
    return Enum(
        ctx,
        line=2,
        box=4,
        staggered=7,
        flank=8
    )

def MyDiplomacyEnum(ctx):
    """Player's Diplomacy Enumeration."""
    return Enum(
        ctx,
        gaia=0,
        self=1,
        ally=2,
        neutral=3,
        enemy=4,
        invalid_player=-1
    )

def ActionEnum(ctx):
    """Action Enumeration."""
    return Enum(
        ctx,
        interact=0,
        stop=1,
        ai_interact=2,
        move=3,
        create=4,
        add_attribute=5,
        give_attribute=6,
        ai_move=10,
        resign=11,
        spec=15,
        waypoint=16,
        stance=18,
        guard=19,
        follow=20,
        patrol=21,
        formation=23,
        save=27,
        ai_waypoint=31,
        chapter=32,
        de_unknown_35=35,
        de_unknown_37=37,
        de_attackmove=33,
        de_autoscout=38,
        de_unknown_39=39,
        ai_command=53,
        ai_queue=100,
        research=101,
        build=102,
        game=103,
        wall=105,
        delete=106,
        attackground=107,
        tribute=108,
        de_unknown_109=109,
        repair=110,
        release=111,
        multiqueue=112,
        togglegate=114,
        flare=115,
        order=117,
        queue=119,
        gatherpoint=120,
        sell=122,
        buy=123,
        droprelic=126,
        townbell=127,
        backtowork=128,
        de_queue=129,
        de_unknown_130=130,
        de_unknown_131=131,
        de_unknown_134=134,
        de_unknown_135=135,
        de_unknown_137=137,
        de_unknown_196=196,
        postgame=255,
        default=Pass
    )



"""Summary compatibility."""
from collections import defaultdict
from mgz.model import parse_match
from mgz.common.diplomacy import get_diplomacy_type
from mgz.common.chat import Chat as ChatType


TC_IDS = [71, 109, 141, 142]
STONE_WALL_ID = 117
PALISADE_WALL_ID = 72


def empty_achievements():
    ach = {}
    ach['military'] = dict(
        score=None,
        units_killed=None,
        units_lost=None,
        buildings_lost=None,
        buildings_razed=None,
        units_converted=None,
        hit_points_killed=None,
        hit_points_razed=None,
    )
    ach['economy'] = dict(
        score=None,
        food_collected=None,
        wood_collected=None,
        stone_collected=None,
        gold_collected=None,
        tribute_sent=None,
        tribute_received=None,
        trade_gold=None,
        relic_gold=None,
    )
    ach['society'] = dict(
        score=None,
        total_relics=None,
        total_castles=None,
        total_wonders=None,
        villager_high=None,
    )
    ach['technology'] = dict(
        score=None,
        explored_percent=None,
        research_count=None,
        research_percent=None,
        feudal_time=None,
        castle_time=None,
        imperial_time=None,
    )
    return ach


class ModelSummary:
    """Compatibility layer between Model and Summary classes."""

    def __init__(self, handle):
        self.match = parse_match(handle)
        self.size = self.match.file.size

    def get_chat(self):
        return [dict(
            type=ChatType.MESSAGE,
            player_number=c.player.number,
            message=c.message,
            timestamp=c.timestamp.total_seconds() * 1000,
            origination=c.origination,
            audience=c.audience
        ) for c in self.match.chat]

    def get_settings(self):
        return dict(
            type=(
                self.match.type_id,
                self.match.type
            ),
            difficulty=(
                self.match.difficulty_id,
                self.match.difficulty
            ),
            population_limit=self.match.population,
            speed=(
                self.match.speed_id,
                self.match.speed
            ),
            cheats=self.match.cheats,
            team_together=self.match.team_together,
            all_technologies=self.match.all_technologies,
            lock_speed=self.match.lock_speed,
            lock_teams=self.match.lock_teams,
            map_reveal_choice=(
                self.match.map_reveal_id,
                self.match.map_reveal
            ),
            diplomacy_type=self.match.diplomacy_type,
            starting_resources=(0, 'Standard'),
            starting_age=(
                self.match.starting_age_id,
                self.match.starting_age
            ),
            ending_age=(3, 'Imperial'),
            victory_condition=(1, 'Conquest'),
            treaty_length=None,
            multiqueue=self.match.multiqueue,
            hidden_civs=self.match.hidden_civs
        )

    def get_file_hash(self):
        return self.match.file.hash

    def get_encoding(self):
        return self.match.file.encoding.name

    def get_hash(self):
        return self.match.hash

    def get_platform(self):
        return dict(
            platform_id='de',
            platform_match_id=self.match.guid,
            ladder=None,
            rated=self.match.rated,
            ratings=None,
            lobby_name=self.match.lobby,
            spec_delay=int(self.match.spec_delay.total_seconds()),
            allow_specs=self.match.allow_specs,
            private=self.match.private
        )

    def get_language(self):
        return self.match.file.language

    def get_device(self):
        return self.match.file.device_type

    def get_owner(self):
        return self.match.file.perspective.number

    def get_duration(self):
        return self.match.duration.total_seconds() * 1000

    def get_completed(self):
        return self.match.completed

    def get_restored(self):
        return self.match.restored, self.match.restored_at.total_seconds() * 1000

    def has_achievements(self):
        return False

    def get_version(self):
        return (
            self.match.version,
            self.match.game_version,
            self.match.save_version,
            self.match.log_version,
            self.match.build_version
        )

    def get_played(self):
        return self.match.timestamp.timestamp() if self.match.timestamp else None

    def get_postgame(self):
        return None

    def get_dataset(self):
        if self.match.dataset_id == 101:
            return dict(
                id=101,
                name='Return of Rome',
                version=None
            )
        return dict(
            id=100,
            name='Definitive Edition',
            version=None
        )

    def get_teams(self):
        return [[p.number for p in t] for t in self.match.teams]

    def get_diplomacy(self):
        d_type = get_diplomacy_type(self.match.teams, self.match.players)
        team_sizes = sorted([len(team) for team in self.match.teams])
        ts = 'v'.join([str(size) for size in team_sizes])
        if d_type == 'FFA':
            ts = 'FFA'
        return dict(
            type=d_type,
            team_size=ts
        )

    def get_players(self):
        return [
            dict(
                name=p.name,
                number=p.number,
                civilization=p.civilization_id,
                color_id=p.color_id,
                human=True,
                winner=p.winner,
                user_id=p.profile_id,
                position=(p.position.x, p.position.y),
                mvp=None,
                score=None,
                rate_snapshot=p.rate_snapshot,
                cheater=None,
                achievements=empty_achievements(),
                prefer_random=p.prefer_random,
                eapm=p.eapm
            ) for p in self.match.players
        ]

    def get_mirror(self):
        mirror = False
        if self.get_diplomacy()['type'] == '1v1':
            civs = set()
            for data in self.get_players():
                civs.add(data['civilization'])
            mirror = (len(civs) == 1)
        return mirror

    def get_objects(self):
        output = []
        tcs = defaultdict(int)
        stone_walls = {}
        palisade_walls = {}
        objects = [(None, self.match.gaia)]
        for player in self.match.players:
            objects.append((player.number, player.objects))
        for player_number, objs in objects:
            for obj in objs:
                if obj.class_id not in [10, 20, 30, 70, 80]:
                    continue
                if obj.index == 1:
                    continue
                if obj.object_id in TC_IDS:
                    tcs[player_number] += 1
                if obj.object_id == STONE_WALL_ID:
                    stone_walls[player_number] = True
                if obj.object_id == PALISADE_WALL_ID:
                    palisade_walls[player_number] = True
                output.append(dict(
                    object_id=obj.object_id,
                    instance_id=obj.instance_id,
                    class_id=obj.class_id,
                    player_number=player_number,
                    x=obj.position.x,
                    y=obj.position.y
                ))
        return dict(
            objects=output,
            tcs=max(tcs.values()) if len(tcs.values()) > 0 else None,
            stone_walls=bool(stone_walls) and all(stone_walls.values()),
            palisade_walls=bool(palisade_walls) and all(palisade_walls.values())
        )

    def get_map(self):
        return dict(
            id=self.match.map.id if not self.match.map.custom else None,
            name=self.match.map.name,
            size=self.match.map.size,
            dimension=self.match.map.dimension,
            custom=self.match.map.custom,
            seed=self.match.map.seed,
            mod_id=self.match.map.mod_id,
            modes=self.match.map.modes,
            zr=self.match.map.zr,
            water=None,
            tiles=[
                dict(
                    x=t.position.x,
                    y=t.position.y,
                    elevation=t.elevation,
                    terrain_id=t.terrain
                ) for t in self.match.map.tiles
            ]
        )



import glob
import unittest
from mgz import header, body, fast
from mgz.summary import ModelSummary, FullSummary


def parse_file_full(path):
    with open(path, 'rb') as f:
        f.seek(0, 2)
        eof = f.tell()
        f.seek(0)
        h = header.parse_stream(f)
        body.meta.parse_stream(f)
        while f.tell() < eof:
            body.operation.parse_stream(f)


def parse_file_fast(path):
    with open(path, 'rb') as f:
        f.seek(0, 2)
        eof = f.tell()
        f.seek(0)
        h = header.parse_stream(f)
        fast.meta(f)
        while f.tell() < eof:
            fast.operation(f)

def parse_file_summary(path, summary_class):
    with open(path, 'rb') as f:
        summary_class(f)


class TestFiles(unittest.TestCase):

    def test_files_full(self):
        parse_file_full('tests/recs/small.mgz')
        parse_file_full('tests/recs/de-13.07.aoe2record')

    def test_files_fast(self):
        # these files aren't supported by full header parser for now:
        skip = {"tests/recs/de-50.6-scenario.aoe2record", "tests/recs/de-50.6-scenario-with-triggers.aoe2record"}

        for path in glob.glob('tests/recs/*'):
            if path.replace("\\", "/") in skip:
                continue
            parse_file_fast(path)

    @unittest.skip("This test is meant to be run manually when debugging issues in a specific rec")
    def test_single_rec(self):
        rec = "tests/recs/de-64.3.aoe2record"
        parse_file_fast(rec)
        parse_file_summary(rec, FullSummary)
        parse_file_summary(rec, ModelSummary)




"""Refine player actions into inputs."""

from mgz.fast import Action as ActionEnum
from mgz.model.definitions import Input


ACTION_TRANSLATE = {
    ActionEnum.DE_QUEUE: 'Queue',
    ActionEnum.DE_ATTACK_MOVE: 'Attack Move'
}


class Inputs:
    """Normalize player inputs."""

    def __init__(self, gaia):
        """Initialize."""
        self._gaia = gaia
        self._buildings = {}
        self._oid_cache = {}
        self.inputs = []

    def add_chat(self, chat):
        """Add chat input."""
        self.inputs.append(Input(chat.timestamp, 'Chat', None, dict(message=chat.message), chat.player, None))

    def add_action(self, action):
        """Add action input."""
        if action.type in (ActionEnum.DE_TRANSFORM, ActionEnum.POSTGAME):
            return
        name = ACTION_TRANSLATE.get(action.type, action.type.name).replace('_', ' ').title()
        param = None
        if 'object_ids' in action.payload and action.payload['object_ids']:
            self._oid_cache[action.type] = action.payload['object_ids']
        elif action.type in self._oid_cache:
            action.payload['object_ids'] = self._oid_cache[action.type]
        if action.type is ActionEnum.SPECIAL:
            name = action.payload['order']
        elif action.type is ActionEnum.GAME:
            name = action.payload['command']
            if name == 'Speed':
                param = action.payload['speed']
        elif action.type is ActionEnum.STANCE:
            name = 'Stance'
            param = action.payload['stance']
        elif action.type is ActionEnum.FORMATION:
            name = 'Formation'
            param = action.payload['formation']
        elif action.type is ActionEnum.ORDER and action.payload['target_id'] in self._gaia:
            name = 'Gather'
            param = self._gaia[action.payload['target_id']]
        elif action.type is ActionEnum.ORDER and action.position and action.position.hash() in self._buildings:
            name = 'Target'
            param = self._buildings[action.position.hash()]
        elif action.type is ActionEnum.GATHER_POINT:
            if action.payload['target_id'] in self._gaia:
                param = self._gaia[action.payload['target_id']]
            elif action.position and action.position.hash() in self._buildings:
                if len(action.payload['object_ids']) == 1 and action.payload['object_ids'][0] == action.payload['target_id']:
                    name = 'Spawn'
                param = self._buildings[action.position.hash()]
        elif action.type in (ActionEnum.BUY, ActionEnum.SELL):
            action.payload['amount'] *= 100
        elif action.type is ActionEnum.BUILD:
            param = action.payload['building']
            if action.position.hash() in self._buildings:
                if self._buildings[action.position.hash()] == 'Farm' and action.payload['building'] == 'Farm':
                    name = 'Reseed'
            self._buildings[action.position.hash()] = action.payload['building']
        elif action.type in (ActionEnum.QUEUE, ActionEnum.DE_QUEUE):
            param = action.payload['unit']
        elif action.type is ActionEnum.RESEARCH:
            param = action.payload['technology']
        new_input = Input(
            action.timestamp,
            name,
            param,
            action.payload,
            action.player,
            action.position
        )
        self.inputs.append(new_input)
        return new_input



"""Determine teams."""
from collections import defaultdict

def get_teams_data(header):
    """Get teams."""
    if header.de:
        # 0: empty slot
        # 1: no team
        # n: n - 1 = team number
        by_team = defaultdict(list)
        for id, player in enumerate(header.de.players):
            if player.resolved_team_id > 1:
                by_team[player.resolved_team_id].append(id + 1)
            elif player.resolved_team_id == 1:
                by_team[id + 9].append(id + 1)
        return set([frozenset(s) for s in by_team.values()])
    allies = {}
    for number, player in enumerate(header.initial.players[1:]):
        allies[number + 1] = set([number + 1])
        for i, mode in enumerate(player.attributes.my_diplomacy):
            if mode == 'ally':
                allies[number + 1].add(i)
    return set([frozenset(s) for s in allies.values()])



"""Reference data helpers."""

import json
import pkgutil
from mgz.util import Version


REF_PACKAGE = 'aocref'


def get_dataset(version, mod):
    """Fetch dataset reference data."""
    if version is Version.DE:
        if isinstance(mod, list) and 11 in mod:
            dataset_id = 101
        else:
            dataset_id = 100
    elif version is Version.HD:
        dataset_id = 300
    elif mod:
        dataset_id = mod[0]
    else:
        dataset_id = 0
    return dataset_id, json.loads(pkgutil.get_data(REF_PACKAGE, f'data/datasets/{dataset_id}.json'))


def get_consts():
    """Fetch constants."""
    return json.loads(pkgutil.get_data(REF_PACKAGE, f'data/constants.json'))



"""Map info."""

from construct import (Array, Byte, Computed, Embedded, Flag, IfThenElse,
                       Int32ul, Padding, Struct, Int16sl, If, Peek)

from mgz.util import Version, find_save_version

# pylint: disable=invalid-name, bad-continuation


# Represents a single map tile, defined by terrain type and elevation.
tile = "tile"/Struct(
    "terrain_type"/Byte,
    Embedded(IfThenElse(lambda ctx: ctx._._.version == Version.DE,
        Embedded(Struct(
            If(lambda ctx: ctx._._._.save_version >= 62.0, Byte),
            Padding(1), # copy of previous byte
            "elevation"/Byte,
            "unk0"/Int16sl,
            "unk1"/Int16sl,
            If(lambda ctx: ctx._._._.save_version >= 13.03 or ctx._._.check.val > 1000, "unk2"/Int16sl)
        )),
        Embedded(IfThenElse(lambda ctx: ctx.terrain_type == 255, Struct(
            "terrain_type"/Byte,
            "elevation"/Byte,
            Padding(1)
        ), Struct(
            "elevation"/Byte
        )))
    ))
)


# Map size and terrain.
map_info = "map_info"/Struct(
    "size_x"/Int32ul,
    "size_y"/Int32ul,
    "tile_num"/Computed(lambda ctx: ctx.size_x * ctx.size_y),
    "zone_num"/Int32ul,
    Array(lambda ctx: ctx.zone_num, Struct(
        IfThenElse(lambda ctx: ctx._._.version in (Version.DE, Version.HD),
            Padding(lambda ctx: 2048 + (ctx._.tile_num * 2)),
            Padding(lambda ctx: 1275 + ctx._.tile_num)
        ),
        "num_floats"/Int32ul,
        Padding(lambda ctx: ctx.num_floats * 4),
        Padding(4)
    )),
    "all_visible"/Flag,
    "fog_of_war"/Flag,
    "check"/Peek(Struct(Padding(lambda ctx: ctx._.tile_num * 7), "val"/Int32ul)), # DE 12.97 fix
    Array(lambda ctx: ctx.tile_num, tile),
    "num_data"/Int32ul,
    Padding(4),
    Array(lambda ctx: ctx.num_data, Padding(4)),
    Array(lambda ctx: ctx.num_data, "couple"/Struct(
        "num_obstructions"/Int32ul,
        Array(lambda ctx: ctx.num_obstructions, Padding(8))
    )),
    "size_x_2"/Int32ul,
    "size_y_2"/Int32ul,
    Padding(lambda ctx: ctx.tile_num * 4), # visibility
    If(lambda ctx: find_save_version(ctx) >= 61.5, Padding(lambda ctx: ctx.tile_num * 4))
)



"""Model class definitions."""

from dataclasses import dataclass
from datetime import timedelta, datetime
from mgz.fast import Action as ActionEnum
from mgz.fast import Age as AgeEnum
from mgz.util import Version


@dataclass
class Position:
    """Represents a coordinate."""

    x: float
    y: float

    def hash(self):
        return hash((self.x, self.y))


@dataclass
class Object:
    """Represents an object."""

    name: str
    class_id: int
    object_id: int
    instance_id: int
    index: int
    position: Position


@dataclass
class TimeseriesRow:
    """Represents a timeseries row."""

    timestamp: timedelta
    total_resources: int
    total_objects: int


@dataclass
class Player:
    """Represents a player."""

    number: int
    name: str
    color: str
    color_id: int
    civilization: str
    civilization_id: int
    position: Position
    objects: list
    profile_id: int
    timeseries: list[TimeseriesRow]
    prefer_random: bool = None
    team: list = None
    team_id: int = None
    winner: bool = False
    eapm: int = None
    rate_snapshot: int = None

    def __repr__(self):
        return self.name

    def __hash__(self):
        return self.number


@dataclass
class Action:
    """Represents an abstract action."""

    timestamp: timedelta
    type: ActionEnum
    payload: dict
    player: Player = None
    position: Position = None


@dataclass
class Input:
    """Represents a player input."""

    timestamp: timedelta
    type: str
    param: str
    payload: dict
    player: Player = None
    position: Position = None


@dataclass
class Viewlock:
    """Represents player view."""

    timestamp: timedelta
    position: Position
    player: Player


@dataclass
class Tile:
    """Represents a map tile."""

    terrain: int
    elevation: int
    position: Position


@dataclass
class Map:
    """Represents a map."""

    id: int
    name: str
    dimension: int
    size: str
    custom: bool
    seed: int
    mod_id: int
    zr: bool
    modes: dict
    tiles: list

    def __repr__(self):
        return self.name

@dataclass
class File:
    """Represents the recorded game file."""

    encoding: str
    language: str
    hash: str
    size: int
    device_type: int
    perspective: Player
    viewlocks: list


@dataclass
class Chat:
    """Represents a chat message."""

    timestamp: timedelta
    message: str
    origination: str
    audience: str
    player: Player

    def __repr__(self):
        return f'[{self.timestamp}] {self.player}: {self.message}'

@dataclass
class Uptime:
    """Represents an advanced to age event."""

    timestamp: timedelta
    age: AgeEnum
    player: Player

    def __repr__(self):
        return f'[{self.timestamp}] {self.player} -> {self.age}'


@dataclass
class Match:
    """Represents a match."""

    players: list
    teams: list
    gaia: list
    map: Map
    file: File
    restored: bool
    restored_at: timedelta
    speed: str
    speed_id: int
    cheats: bool
    lock_teams: bool
    population: int
    chat: list
    guid: str
    lobby: str
    rated: bool
    dataset: str
    type: str
    type_id: int
    map_reveal: str
    map_reveal_id: int
    difficulty: str
    difficulty_id: int
    starting_age: str
    starting_age_id: int
    team_together: bool
    lock_speed: bool
    all_technologies: bool
    multiqueue: bool
    duration: timedelta
    diplomacy_type: str
    completed: bool
    dataset_id: int
    version: Version
    game_version: str
    save_version: float
    log_version: int
    build_version: int
    timestamp: datetime
    spec_delay: timedelta
    allow_specs: bool
    hidden_civs: bool
    private: bool
    hash: str
    actions: list
    inputs: list
    uptimes: list



"""Entry point for reading MGZ."""

# pylint: disable=invalid-name,no-name-in-module

from construct import (Struct, CString, Const, Int32ul, Embedded, Float32l, Terminated, If, Computed, this, Peek, Bytes, Int32ub)
from mgz.util import MgzPrefixed, ZlibCompressed, Version, VersionAdapter, get_version, get_save_version
from mgz.header.ai import ai
from mgz.header.replay import replay
from mgz.header.map_info import map_info
from mgz.header.initial import initial
from mgz.header.achievements import achievements
from mgz.header.scenario import scenario
from mgz.header.lobby import lobby
from mgz.header.de import de
from mgz.header.hd import hd


compressed_header = Struct(
    "game_version"/CString(encoding='latin1'),
    "checker"/Peek(Float32l),
    "old_save_version"/VersionAdapter(Float32l),
    "new_save_version"/If(lambda ctx: ctx.old_save_version == -1, Int32ul),
    "save_version"/Computed(lambda ctx: get_save_version(ctx.old_save_version, ctx.new_save_version)),
    "version"/Computed(lambda ctx: get_version(ctx.game_version, ctx.save_version, None)),
    "hd"/If(lambda ctx: ctx.version == Version.HD and ctx.save_version > 12.34, hd),
    "de"/If(lambda ctx: ctx.version == Version.DE, de),
    ai,
    replay,
    map_info,
    initial,
    achievements,
    scenario,
    lobby,
    Terminated
)


subheader = Struct(
    "check"/Peek(Int32ul),
    "chapter_address"/If(lambda ctx: ctx.check < 100000000, Int32ul),
    Embedded(MgzPrefixed(lambda ctx: ctx._.header_length - 4 - (4 if ctx.check < 100000000 else 0), ZlibCompressed(compressed_header)))
)

"""Header is compressed"""
header = Struct(
    "header_length"/Int32ul,
    Embedded(subheader),
    "log_version"/If(lambda ctx: ctx.save_version >= 11.76, Peek(Int32ul)),
    "version"/Computed(lambda ctx: get_version(ctx.game_version, ctx.save_version, ctx.log_version))
)



"""Body.

An mgz body is a stream of Operations

An Operation can be:
 - Action: Player input that materially affects the game
 - Chat: Player chat message
 - Synchronization: Time increment
 - Viewlock: View coordinates of recording player
 - Embedded: A variety of embedded structures
"""

from construct import (Struct, Byte, Switch, Embedded, Padding,
                       Int32ul, Peek, Tell, Float32l, String, If, Array, Bytes,
                       GreedyBytes, Computed, IfThenElse, Int16ul, Int64ul, Seek, Padded)
from mgz.enums import ActionEnum, OperationEnum
from mgz.body import actions, embedded
from mgz.util import BoolAdapter


# pylint: disable=invalid-name, bad-continuation


# Action.
action_data = "action"/Struct(
    Peek("type_int"/Byte),
    ActionEnum("type"/Byte),
    Embedded("action"/Switch(lambda ctx: ctx.type, {
        "interact": actions.interact,
        "move": actions.move,
        "stop": actions.stop,
        "create": actions.create,
        "stance": actions.stance,
        "guard": actions.guard,
        "follow": actions.follow,
        "formation": actions.formation,
        "waypoint": actions.waypoint,
        "give_attribute": actions.give_attribute,
        "add_attribute": actions.add_attribute,
        "ai_waypoint": actions.ai_waypoint,
        "ai_interact": actions.ai_interact,
        "ai_move": actions.ai_move,
        "ai_queue": actions.ai_queue,
        "save": actions.save,
        "chapter": actions.chapter,
        "ai_command": actions.ai_command,
        "spec": actions.spec,
        "build": actions.build,
        "game": actions.game,
        "patrol": actions.patrol,
        "wall": actions.wall,
        "delete": actions.delete,
        "attackground": actions.attackground,
        "repair": actions.repair,
        "release": actions.release,
        "togglegate": actions.togglegate,
        "flare": actions.flare,
        "order": actions.order,
        "droprelic": actions.droprelic,
        "gatherpoint": actions.gatherpoint,
        "townbell": actions.townbell,
        "resign": actions.resign,
        "tribute": actions.tribute,
        "queue": actions.queue,
        "multiqueue": actions.multiqueue,
        "research": actions.research,
        "sell": actions.sell,
        "buy": actions.buy,
        "backtowork": actions.backtowork,
        "de_queue": actions.de_queue,
        "postgame": actions.postgame,
        "de_attackmove": actions.de_attackmove,
        "de_autoscout": actions.de_autoscout
    }, default=Struct(
        "unk_action"/Computed(lambda ctx: ctx._.type),
        "bytes"/Bytes(lambda ctx: ctx._._.length - 1)
    ))),
    Padding(4)
)


# Action - length followed by data.
action = "action"/Struct(
    "length"/Int32ul,
    Padded(lambda ctx: ctx.length + 4, action_data)
)


# Synchronization.
sync = "sync"/Struct(
    "time_increment"/Int32ul,
    Peek("next"/Int32ul),
    "checksum"/If(lambda ctx: not ctx.next, Struct(
        Padding(8),
        "sync"/Int32ul,
        "unknown"/Bytes(4),
        "sequence"/Int32ul,
        If(lambda ctx: ctx.sequence > 0, Bytes(332)),
        Bytes(8)
    ))
)


# View lock.
viewlock = "viewlock"/Struct(
    "view"/Struct(
        "x"/Float32l,
        "y"/Float32l
    ),
    "player_id"/Int32ul
)


# Chat.
chat = "chat"/Struct(
    Padding(4),
    "length"/Int32ul,
    "text"/Bytes(lambda ctx: ctx.length - 0),
    Padding(0)
)


# Start.
start = "start"/Struct(
    Seek(-4, whence=1),
    "checksum_interval"/Int32ul,
    BoolAdapter("multiplayer"/Int32ul),
    "rec_owner"/Int32ul,
    BoolAdapter("reveal_map"/Int32ul),
    "use_sequence_numbers"/Int32ul,
    "number_of_chapters"/Int32ul,
    "aok_or_de"/Peek(Int32ul),
    If(lambda ctx: ctx.aok_or_de == 0, Struct(
        Int32ul,
        "aok"/Peek(Int32ul),
        If(lambda ctx: ctx.aok != 2, Int32ul)
    ))
)

# Meta.
meta = "meta"/Struct(
    "next"/Peek(Int32ul),
    "log_version"/If(lambda ctx: ctx.next != 500, Int32ul),
    "checksum_interval"/Int32ul,
    BoolAdapter("multiplayer"/Int32ul),
    "rec_owner"/Int32ul,
    BoolAdapter("reveal_map"/Int32ul),
    "use_sequence_numbers"/Int32ul,
    "number_of_chapters"/Int32ul,
    "aok_or_de"/Peek(Int32ul),
    "m"/If(lambda ctx: ctx.aok_or_de == 0, Struct(
        "v"/Int32ul,
        "aok"/Peek(Int32ul),
        "z"/If(lambda ctx: ctx.aok != 2, Int64ul)
    ))
)


# Operation.
operation = "operation"/Struct(
    Peek(OperationEnum("type"/Int32ul)),
    "start"/Tell,
    "op"/Int32ul,
    Embedded("data"/Switch(lambda ctx: ctx.type, {
        "action": action,
        "sync": sync,
        "viewlock": viewlock,
        "chat": chat,
        "embedded": embedded.embedded
    })),
    "end"/Tell
)



"""Determine match settings."""


import mgz


def _get_all_techs(postgame, de_data):
    """Get all techs flag."""
    if de_data is not None:
        return de_data.all_techs
    if postgame is not None:
        return postgame.all_techs
    return None


def _get_lock_speed(postgame, de_data):
    """Get lock speed flag."""
    if de_data is not None:
        return de_data.lock_speed
    if postgame is not None:
        return postgame.lock_speed
    return None


def _get_team_together(postgame, de_data):
    """Get team together flag."""
    if de_data is not None:
        return not de_data.random_positions
    if postgame is not None:
        return not postgame.random_positions
    return None


def _get_victory_type(postgame, de_data):
    """Get victory type."""
    if de_data is not None:
        return (de_data.victory_type_id, de_data.victory_type)
    if postgame is not None:
        return (postgame.victory_type_id, postgame.victory_type)
    return (None, None)


def _get_starting_resources(postgame, de_data):
    """Get starting resources."""
    if de_data is not None:
        return (de_data.starting_resources_id, de_data.starting_resources)
    if postgame is not None:
        return (postgame.starting_resources_id, postgame.starting_resources)
    return (None, None)


def _get_starting_age(postgame, de_data):
    """Get starting age."""
    if de_data is not None:
        return (de_data.starting_age_id, de_data.starting_age)
    if postgame is not None:
        return (postgame.starting_age_id, postgame.starting_age)
    return (None, None)


def _get_cheats(header):
    """Get cheat mode."""
    if header.de is not None:
        return header.de.cheats
    if hasattr(header.replay, 'cheats_enabled'):
        return header.replay.cheats_enabled
    return False


def get_settings_data(postgame, header):
    """Get settings."""
    population_limit = header.lobby.population_limit
    game_speed_id = int(header.replay.game_speed_float * 100)
    return {
        'type': (
            header.lobby.game_type_id,
            header.lobby.game_type
        ) if hasattr(header.lobby, 'game_type_id') else (0, 'RM'),
        'difficulty': (
            header.scenario.game_settings.difficulty_id,
            header.scenario.game_settings.difficulty
        ),
        'population_limit': population_limit,
        'map_reveal_choice': (
            header.lobby.reveal_map_id,
            header.lobby.reveal_map
        ),
        'speed': (
            game_speed_id,
            mgz.const.SPEEDS.get(game_speed_id)
        ),
        'starting_resources': _get_starting_resources(postgame, header.de),
        'starting_age': _get_starting_age(postgame, header.de),
        'ending_age': (
            header.de.ending_age_id,
            header.de.ending_age
        ) if header.de else (None, None),
        'victory_condition': _get_victory_type(postgame, header.de),
        'treaty_length': header.de.treaty_length if header.de else None,
        'cheats': _get_cheats(header),
        'team_together': _get_team_together(postgame, header.de),
        'all_technologies': _get_all_techs(postgame, header.de),
        'lock_speed': _get_lock_speed(postgame, header.de),
        'lock_teams': header.lobby.lock_teams if hasattr(header.replay, 'lock_teams') else True,
        'multiqueue': True if header.de is not None else None,
        'hidden_civs': header.de.hidden_civs if header.de is not None else None
    }



"""MGZ parsing utilities."""

import logging
import re
import struct
import zlib
from enum import Enum
from io import BytesIO

import construct.core
from construct import Adapter, Construct, Subconstruct, Tunnel

from mgz import const


# pylint: disable=abstract-method,protected-access

LOGGER = logging.getLogger(__name__)
SEARCH_MAX_BYTES = 3000
POSTGAME_LENGTH = 2096
LOOKAHEAD = 9


class Version(Enum):
    """Version enumeration.

    Using consts from https://github.com/goto-bus-stop/recanalyst/blob/master/src/Model/Version.php
    for consistency.
    """
    AOK = 1
    AOC = 4
    AOC10 = 5
    AOC10C = 8
    USERPATCH12 = 12
    USERPATCH13 = 13
    USERPATCH14 = 11
    USERPATCH15 = 20
    DE = 21
    USERPATCH14RC2 = 22
    MCP = 30
    HD = 19


class MgzPrefixed(Subconstruct):
    """Like `Prefixed`, but accepting arbitrary length."""

    __slots__ = ["name", "length", "subcon"]

    def __init__(self, length, subcon):
        """Initialize."""
        super(MgzPrefixed, self).__init__(subcon)
        self.length = length

    def _parse(self, stream, context, path):
        """Parse tunnel."""
        length = self.length(context)
        new_stream = BytesIO(construct.core._read_stream(stream, length))
        return self.subcon._parse(new_stream, context, path)


class ZlibCompressed(Tunnel):
    """Like Compressed, but only does header-less zlib."""

    __slots__ = []

    def _decode(self, data, context):
        """Decode zlib without header bytes."""
        return zlib.decompress(data, wbits=-15)


def get_save_version(old_version, new_version):
    """Get the save version."""
    if old_version == -1:
        if new_version == 37:
            return 37.0
        return round(new_version / (1<<16), 2)
    return old_version


def get_version(game_version, save_version, log_version):
    """Get version based on version fields."""
    if game_version == 'VER 9.3':
        return Version.AOK
    if game_version == 'VER 9.4':
        if log_version == 3:
            return Version.AOC10
        if log_version == 5 or save_version >= 12.97:
            return Version.DE
        if save_version >= 12.36:
            return Version.HD
        if log_version == 4:
            return Version.AOC10C
        return Version.AOC
    if game_version == 'VER 9.8':
        return Version.USERPATCH12
    if game_version == 'VER 9.9':
        return Version.USERPATCH13
    if game_version == 'VER 9.A':
        return Version.USERPATCH14RC2
    if game_version in ['VER 9.B', 'VER 9.C', 'VER 9.D']:
        return Version.USERPATCH14
    if game_version in ['VER 9.E', 'VER 9.F']:
        return Version.USERPATCH15
    if game_version == 'MCP 9.F':
        return Version.MCP
    if log_version is not None or game_version != 'VER 9.4':
        raise ValueError('unsupported version: {}, {}, {}'.format(game_version, save_version, log_version))


def find_version(ctx):
    """Find version."""
    if 'version' not in ctx:
        return find_version(ctx._)
    return ctx.version


def find_save_version(ctx):
    """Find save version."""
    if 'save_version' not in ctx:
        return find_save_version(ctx._)
    return ctx.save_version


def find_type(ctx):
    """Find object type."""
    if 'type' not in ctx:
        return find_type(ctx._)
    return ctx.type


def check_flags(peek):
    """Check byte sequence for only flag bytes."""
    for i in peek:
        if i not in [0, 1]:
            return False
    return True


def convert_to_timestamp(time):
    """Convert int to timestamp string."""
    if time == -1:
        return None
    time = int(time*1000)
    hour = time//1000//3600
    minute = (time//1000//60) % 60
    second = (time//1000) % 60
    return str(hour).zfill(2)+":"+str(minute).zfill(2)+":"+str(second).zfill(2)


class TimeSecAdapter(Adapter):
    """Conversion to readable time."""

    def _decode(self, obj, context):
        """Decode timestamp to string."""
        return convert_to_timestamp(obj)


class BoolAdapter(Adapter):
    """Bools with potential padding."""

    def _decode(self, obj, context):
        """Decode bool."""
        return obj == 1


class VersionAdapter(Adapter):
    """Round save version."""

    def _decode(self, obj, context):
        """Decode by rounding float."""
        return round(obj, 2)


class ModVersionAdapter(Adapter):
    """Parse mod version."""

    def _decode(self, obj, context):
        """Decode mod."""
        number = int(obj)
        mod_id = int(number / 1000)
        mod_version = '.'.join(list(str(number % 1000)))
        return {
            'id': mod_id,
            'name': const.MODS.get(mod_id, 'Unknown Mod'),
            'version': mod_version
        }


class Find(Construct):
    """Find bytes, and read past them."""

    __slots__ = ["find", "max_length"]

    def __init__(self, find, max_length):
        """Initiallize."""
        Construct.__init__(self)
        if not isinstance(find, list):
            find = [find]
        self.find = find
        self.max_length = max_length

    def _parse(self, stream, context, path):
        """Parse stream to find a given byte string."""
        start = stream.tell()
        read_bytes = ""
        if self.max_length:
            read_bytes = stream.read(self.max_length)
        else:
            read_bytes = stream.read()
        candidates = []
        for f in self.find:
            match = re.search(f, read_bytes, re.DOTALL)
            if not match:
                continue
            candidates.append(match.end())
        if not candidates:
            raise RuntimeError('could not find bytes: {}'.format(f))
        choice = min(candidates)
        stream.seek(start + choice)
        return choice


class RepeatUpTo(Subconstruct):
    """Like RepeatUntil, but doesn't include the last element in the return value."""

    __slots__ = ["find"]

    def __init__(self, find, subcon):
        """Initialize."""
        Subconstruct.__init__(self, subcon)
        self.find = find

    def _parse(self, stream, context, path):
        """Parse until a given byte string is found."""
        objs = []
        while True:
            start = stream.tell()
            test = stream.read(len(self.find))
            stream.seek(start)
            if test == self.find:
                break
            subobj = self.subcon._parse(stream, context, path)
            objs.append(subobj)
        return objs


class GotoObjectsEnd(Construct):
    """Find the end of a player's objects list.

    Necessary since we can't parse objects from a resume game (yet).
    """

    # pylint: disable=chained-comparison, too-many-locals
    def _parse(self, stream, context, path):
        """Parse until the end of objects data."""
        num_players = context._._.replay.num_players
        marker_num = context.attributes.num_header_data
        save_version = context._._.save_version
        version = find_version(context)
        start = stream.tell()
        # Have to read everything to be able to use find()
        read_bytes = stream.read()
        # Try to find the first marker, a portion of the next player structure
        # The byte that changes is the number of player stats fields
        marker = read_bytes.find(b'\x16' + struct.pack('<I', int(marker_num)) + b'\x21')
        # If it exists, we're not on the last player yet
        if marker > 0:
            # Backtrack through the player name
            count = 0
            while struct.unpack("<H", read_bytes[marker-2:marker])[0] != count:
                marker -= 1
                count += 1
            # Backtrack through the rest of the next player structure
            offset = 9 * 4
            if save_version >= 61.5:
                offset = num_players * 4
            backtrack = 7 + num_players + offset
        # Otherwise, this is the last player
        else:
            # Search for the scenario header
            for i, b in enumerate(read_bytes):
                if b == 63 and b"\xff\xff\xff\xff\x00\x00\x00\x00" == read_bytes[i-11:i-3]:
                    flt = struct.unpack("<f", read_bytes[i-3:i+1])[0]
                    # scenario header version is something like 1.x
                    if flt > 1.0 and flt < 2.0:
                        marker = i - 3
                        break
            # Backtrack through the achievements and initial structure footer
            backtrack = ((1817 * (num_players - 1)) + 4 + 19)
        # Seek to the position we found
        end = start + marker - backtrack - 2
        stream.seek(end)
        return end


def find_postgame(data, size):
    """Find postgame and grab duration.

    We can find postgame location by scanning the last few
    thousand bytes of the rec and looking for a pattern as
    follows:

    [action op]    [action length]    [action type]
    01 00 00 00    30 08 00 00        ff

    The last occurance of this pattern signals the start of
    the postgame structure. Note that the postgame action length
    is always constant, unlike other actions.
    """
    pos = None
    for i in range(size - SEARCH_MAX_BYTES, size - LOOKAHEAD):
        op_type, length, action_type = struct.unpack('<IIB', data[i:i + LOOKAHEAD])
        if op_type == 0x01 and length == POSTGAME_LENGTH and action_type == 0xFF:
            LOGGER.debug("found postgame candidate @ %d with length %d", pos, length)
            return i + LOOKAHEAD, length
    return None, None


def unpack(fmt, data, shorten=True):
    """Unpack bytes according to format string."""
    output = struct.unpack(fmt, data.read(struct.calcsize(fmt)))
    if len(output) == 1 and shorten:
        return output[0]
    return output


def as_hex(d):
    return " ".join(["{:02x}".format(x) for x in d])



"""MGZ tools."""

import asyncio
import argparse
import logging
import os
import struct
import sys
from datetime import datetime
from collections import defaultdict

from construct.core import ConstructError
from tabulate import tabulate

import mgz
import mgz.const
import mgz.header
import mgz.util
from mgz import fast
from mgz.summary import Summary
from mgz.util import find_postgame, LOOKAHEAD


LOGGER = logging.getLogger(__name__)
CMD_INFO = 'info'
CMD_CHAT = 'chat'
CMD_VALIDATE = 'validate'
CMD_DUMP = 'dump'
CMD_MERGE = 'merge'
CMD_HISTOGRAM = 'histogram'
CMD_PAD = 'pad'


def print_info(path):
    """Print basic info."""
    with open(path, 'rb') as handle:
        summary = Summary(handle)
        dataset = summary.get_dataset()
        print('-------------')
        print(tabulate([
            ['Path', path],
            ['Duration', mgz.util.convert_to_timestamp(summary.get_duration() / 1000)],
            ['Played', datetime.utcfromtimestamp(summary.get_played()) if summary.get_played() else None],
            ['Completed', summary.get_completed()],
            ['Restored', summary.get_restored()[0]],
            ['Postgame', bool(summary.get_postgame())],
            ['Version', '{} ({}, {}, {}, {})'.format(*summary.get_version())],
            ['Dataset', '{} {}'.format(dataset['name'], dataset['version'])],
            ['File Hash', summary.get_file_hash()],
            ['Match Hash', summary.get_hash().hexdigest() if summary.get_hash() else None],
            ['Encoding', summary.get_encoding()],
            ['Language', summary.get_language()],
            ['Device', summary.get_device()],
            ['Map', '{} ({}, {})'.format(summary.get_map()['name'], summary.get_map()['seed'], summary.get_map()['mod_id'])] # pylint: disable=unsubscriptable-object
        ], tablefmt='plain'))


def is_valid(path):
    """Validate a recorded game."""
    with open(path, 'rb') as handle:
        size = os.fstat(handle.fileno()).st_size
        try:
            mgz.header.parse_stream(handle)
            mgz.body.meta.parse_stream(handle)
            while handle.tell() < size:
                mgz.body.operation.parse_stream(handle)
            print('valid')
            return True
        except ConstructError:
            print('invalid')
            return False


def dump_rec(path):
    """Write parsed game to stdout."""
    with open(path, 'rb') as handle:
        size = os.fstat(handle.fileno()).st_size
        mgz.header.parse_stream(handle)
        while handle.tell() < size:
            operation = mgz.body.operation.parse_stream(handle)
            if operation.type == 'embedded':
                operation.data = '<snipped>'
            print(operation)


def print_chat(path):
    """Extract chat."""
    with open(path, 'rb') as handle:
        summary = Summary(handle)
        for c in summary.get_chat():
            print(c)


def merge_recs(part_one, part_two, output):
    """Merge two recorded games."""
    start_op_length = 28
    with open(part_one, 'rb') as a_handle, \
        open(part_two, 'rb') as b_handle, \
        open(output, 'wb') as merged:

        a_data = a_handle.read()
        b_data = b_handle.read()

        postgame_pos, _ = find_postgame(a_data, len(a_data))
        if postgame_pos:
            a_data_end = postgame_pos - LOOKAHEAD
        else:
            a_data_end = len(a_data)
        b_header_len, = struct.unpack('<I', b_data[:4])
        chapter = mgz.body.operation.build({
            'type': 'action',
            'op': 1,
            'length': 2,
            'action': {
                'type': 'chapter',
                'player_id': 0xff # our merge marker
            }
        })

        # part A with no postgame struct
        merged.write(a_data[:a_data_end])
        # chapter action
        merged.write(chapter)
        # offset to start of part B operations
        merged.write(struct.pack('<I', a_data_end + len(chapter) + b_header_len))
        # part B header (now a "saved chapter")
        merged.write(b_data[4:b_header_len])
        # part B operations with no start operation
        merged.write(b_data[b_header_len + start_op_length:])


def pad_rec(target_size, path, output):
    """Pad a recorded game."""
    with open(path, 'rb') as handle, open(output, 'wb') as padded:
        data = handle.read()
        data_length = len(data)
        padded.write(data)
        pad_length = target_size - data_length
        if pad_length < 9:
            raise ValueError('target size too small')
        pad_op = struct.pack('<IIB', 1, pad_length, 0xfe)
        pad_op += b'\x00' * (pad_length - 9)
        padded.write(pad_op)


def print_histogram(path):
    """Show operation and action histogram."""
    with open(path, 'rb') as handle:
        size = os.fstat(handle.fileno()).st_size
        mgz.header.parse_stream(handle)
        operations = defaultdict(int)
        actions = defaultdict(int)
        labels = {}
        fast.meta(handle)
        while handle.tell() < size:
            op_type, payload = fast.operation(handle)
            operations[op_type.name] += 1
            if op_type == fast.Operation.ACTION:
                action_id = payload[1]
                action_id = '{0:#0{1}x}'.format(payload[0].value, 4)
                labels[action_id] = payload[0].name
                actions[action_id] += 1
        print('Operations')
        print(tabulate([
            [operation, operations[operation]]
            for operation in sorted(operations, key=operations.get, reverse=True)
        ], headers=['Name', 'Count'], tablefmt='simple'))
        print()
        print('Actions')
        print(tabulate([
            [action, labels[action], actions[action]]
            for action in sorted(actions, key=actions.get, reverse=True)
        ], headers=['ID', 'Name', 'Count'], tablefmt='simple'))


async def run(args): # pylint: disable=too-many-branches
    """Entry point."""
    if args.cmd == CMD_INFO:
        for rec in args.rec_path:
            print_info(rec)
    elif args.cmd == CMD_CHAT:
        for rec in args.rec_path:
            print_chat(rec)
    elif args.cmd == CMD_VALIDATE:
        for rec in args.rec_path:
            if not is_valid(rec):
                sys.exit(1)
    elif args.cmd == CMD_DUMP:
        for rec in args.rec_path:
            dump_rec(rec)
    elif args.cmd == CMD_MERGE:
        merge_recs(args.part_one, args.part_two, args.output)
    elif args.cmd == CMD_HISTOGRAM:
        for rec in args.rec_path:
            print_histogram(rec)
    elif args.cmd == CMD_PAD:
        pad_rec(args.target_size, args.rec_path, args.output)
    await asyncio.sleep(0)


def get_args():
    """Get arguments."""
    parser = argparse.ArgumentParser()
    subparsers = parser.add_subparsers(dest='cmd')
    subparsers.required = True
    info = subparsers.add_parser(CMD_INFO)
    info.add_argument('rec_path', nargs='+')
    chat = subparsers.add_parser(CMD_CHAT)
    chat.add_argument('rec_path', nargs='+')
    validate = subparsers.add_parser(CMD_VALIDATE)
    validate.add_argument('rec_path', nargs='+')
    dump = subparsers.add_parser(CMD_DUMP)
    dump.add_argument('rec_path', nargs='+')
    merge = subparsers.add_parser(CMD_MERGE)
    merge.add_argument('part_one')
    merge.add_argument('part_two')
    merge.add_argument('output', default='merged.mgz')
    histogram = subparsers.add_parser(CMD_HISTOGRAM)
    histogram.add_argument('rec_path', nargs='+')
    pad = subparsers.add_parser(CMD_PAD)
    pad.add_argument('target_size', type=int)
    pad.add_argument('rec_path')
    pad.add_argument('output')
    return parser.parse_args()


def main():
    """Entry point."""
    logging.basicConfig(level='INFO')
    loop = asyncio.get_event_loop()
    loop.run_until_complete(run(get_args()))


if __name__ == '__main__':
    main()



"""Objects."""

from construct import (Array, Byte, Embedded, Flag, Float32l, If, Int16sl, Int16ub,
                       Int16ul, Int32ub, Int32ul, Padding, Peek, Struct, Int32sl,
                       Switch, Pass, RepeatUntil, Bytes, LazyBound, IfThenElse)

from mgz.enums import ObjectEnum, ObjectTypeEnum, ResourceEnum
from mgz.util import Find, Version, find_version, find_save_version, find_type
from mgz.header.unit_type import unit_type
from mgz.header.de import de_string

# pylint: disable=invalid-name


active_sprite = "active_sprite"/Struct(
    "id"/Int16ul,
    "x"/Int32ul,
    "y"/Int32ul,
    "frame"/Int16ul,
    "invisible"/Byte
)

animated_sprite = "animated_sprite"/Struct(
    "animate_interval"/Float32l,
    "animate_last"/Int32ul,
    "last_frame"/Int16ul,
    "frame_changed"/Byte,
    "frame_looped"/Byte,
    "animate_flag"/Byte,
    "last_speed"/Float32l
)

sprite_list = "sprite_list"/Struct(
    "type"/Byte,
    "active"/If(lambda ctx: ctx.type != 0, active_sprite),
    "animated"/If(lambda ctx: ctx.type == 2, animated_sprite),
    Embedded(If(lambda ctx: ctx.type != 0, Struct(
        "order"/Byte,
        "flag"/Byte,
        "count"/Byte
    )))
)

particle = "particle"/Struct(
    "type"/Int16ul,
    Embedded(If(lambda ctx: ctx.type == 1, Struct(
        "name"/de_string,
        "x"/Float32l,
        "y"/Float32l,
        Bytes(26)
    ))),
)


static = "static"/Struct(
    "object_type"/Int16ul,
    "sprite"/Int16ul,
    "garrisoned_in_id"/Int32sl,
    "hitpoints"/Float32l,
    "object_state"/Byte,
    "sleep"/Flag,
    "doppleganger"/Flag,
    "go_to_sleep"/Flag,
    "object_id"/Int32ul,
    "facet"/Byte,
    "x"/Float32l,
    "y"/Float32l,
    "z"/Float32l,
    "screen_offset_x"/Int16ul,
    "screen_offset_y"/Int16ul,
    "shadow_offset_x"/Int16ul,
    "shadow_offset_y"/Int16ul,
    "selected_group"/If(lambda ctx: find_version(ctx) == Version.AOK, Byte),
    ResourceEnum("resource_type"/Int16sl),
    "amount"/Float32l,
    "worker_count"/IfThenElse(lambda ctx: find_save_version(ctx) == 12.36, Int32ul, Byte),
    "current_damage"/Byte,
    "damaged_lately_timer"/Byte,
    "under_attack"/Byte,
    If(lambda ctx: find_save_version(ctx) < 37, Struct(
        "pathing_group_len"/Int32ul,
        "pathing_group"/Array(lambda ctx: ctx.pathing_group_len, "object_id"/Int32ul)
    )),
    If(lambda ctx: find_save_version(ctx) >= 66.3, Bytes(2)),
    "group_id"/Int32sl,
    "roo_already_called"/Byte,
    "de_static_unk1"/If(lambda ctx: find_version(ctx) == Version.DE, Bytes(17)),
    If(lambda ctx: find_save_version(ctx) >= 26.16, Byte),
    "de_has_object_props"/If(lambda ctx: find_version(ctx) == Version.DE, Int16ul),
    "de_object_props"/IfThenElse(lambda ctx: ctx.de_has_object_props == 1, Struct(
        Bytes(162),
        Find(b'`\n\x00\x00`\n\x00\x00', 1000), # Skip a large unparseable block that (likely) contains RMS-modified object data
        Bytes(2)
    ), Struct(
        "has_sprite_list"/Byte,
        "sprite_list"/If(lambda ctx: ctx.has_sprite_list != 0, RepeatUntil(lambda x,lst,ctx: lst[-1].type == 0, sprite_list)),
        "de_extension"/If(lambda ctx: find_version(ctx) == Version.DE, Struct(
            "particles"/Array(5, particle),
            If(lambda ctx: find_save_version(ctx) >= 13.15, Bytes(5)),
            If(lambda ctx: find_save_version(ctx) >= 13.17, Bytes(2)),
            If(lambda ctx: find_save_version(ctx) >= 13.34, Struct(
                Bytes(2),
                de_string,
                de_string,
                Bytes(2)
            ))
        ))
    )),
    "de_extension_p2"/If(lambda ctx: find_version(ctx) == Version.DE, Struct(
        # not the right way to do this, needs improvement
        If(lambda ctx: find_save_version(ctx) >= 20.16, Struct(
            "peek"/Peek(Bytes(6)),
            If(lambda ctx: find_save_version(ctx) >= 25.22 and (find_type(ctx) == 10 or find_save_version(ctx) >= 63.0), Bytes(1)),
            If(lambda ctx: find_save_version(ctx) >= 64.3, Bytes(1)),
            If(lambda ctx: find_save_version(ctx) < 25.22 and find_type(ctx) == 10 and ctx.peek[0] == 0 and ctx.peek[0:2] != b"\x00\x0b", Bytes(1)),
            If(lambda ctx: find_type(ctx) == 20 and ctx.peek[4] == 0 and ctx.peek[4:6] != b"\x00\x0b", Bytes(1)),
        )),
        If(lambda ctx: find_save_version(ctx) >= 61.5, Bytes(14)),
        If(lambda ctx: find_save_version(ctx) >= 62.0, Bytes(4))
    )),
    "hd_extension"/If(lambda ctx: find_version(ctx) == Version.HD and find_save_version(ctx) > 12.36, Struct(
        "flag"/Flag,
        Padding(4),
        "has_array"/Int16ul,
        "array"/If(lambda ctx: ctx.has_array, Struct(
            "values"/RepeatUntil(lambda x,lst,ctx: lst[-1].type == 0, Struct(
                "type"/Byte,
                If(lambda ctx: ctx.type > 0, Bytes(16))
            ))
        ))
    )),
    If(lambda ctx: find_save_version(ctx) >= 66.3, Bytes(4)),
)


animated = "animated"/Struct(
    Embedded(static),
    "turn_speed"/Float32l
)

path_data = "path_data"/Struct(
    "id"/Int32ul,
    "linked_path_type"/Int32ul,
    "waypoint_level"/Int32ul,
    "path_id"/Int32ul,
    "waypoint"/Int32ul,
    "state"/Int32sl,
    "range"/Float32l,
    "target_id"/Int32sl,
    "pause_time"/Float32l,
    "continue_counter"/Int32ul,
    "flags"/Int32ul,
    "hd"/If(lambda ctx: find_version(ctx) == Version.HD, Int32ul)
)

vector = Struct(
    "x"/Float32l,
    "y"/Float32l,
    "z"/Float32l
)


movement_data = "movement"/Struct(
    "velocity"/vector,
    "acceleration"/vector
)

base_moving = "base_moving"/Struct(
    Embedded(animated),
    "trail_remainder"/Int32ul,
    "velocity"/vector,
    "de_move_byte"/If(lambda ctx: find_save_version(ctx) >= 20.16, Byte),
    "angle"/Float32l,
    "turn_towards_time"/Int32ul,
    If(lambda ctx: find_save_version(ctx) < 37, Struct(
        "turn_timer"/Int32ul,
        "continue_counter"/Int32ul,
        "current_terrain_exception_1"/Int32sl,
        "current_terrain_exception_2"/Int32sl,
    )),
    "waiting_to_move"/Byte,
    "wait_delays_count"/Byte,
    "on_ground"/Byte,
    "num_path_data"/Int32ul,
    "path_data"/Array(lambda ctx: ctx.num_path_data, path_data),
    "has_future_path_data"/Int32ul,
    "future_path_data"/If(lambda ctx: ctx.has_future_path_data > 0, path_data),
    "has_movement_data"/Int32ul,
    "movement_data"/If(lambda ctx: ctx.has_movement_data > 0, movement_data),
    "de_moving_unk"/If(lambda ctx: find_version(ctx) == Version.DE and find_save_version(ctx) < 13.2, Int16ul),
    "position"/vector,
    "orientation_forward"/vector,
    "orientation_right"/vector,
    "last_move_time"/Int32ul,
    "num_user_defined_waypoints"/Int32ul,
    "user_defined_waypoints"/Array(lambda ctx: ctx.num_user_defined_waypoints, vector),
    "has_substitute_position"/Int32ul,
    "substitute_position"/vector,
    "consecutive_substitute_count"/Int32ul
)

moving = "moving"/Struct(
    Embedded(base_moving),
    "hd_moving"/If(lambda ctx: find_version(ctx) == Version.HD, Bytes(1)),
    "de_moving"/If(lambda ctx: find_version(ctx) == Version.DE, Bytes(17)),
    "ver2616"/If(lambda ctx: 37 > find_save_version(ctx) >= 26.16, Bytes(8)),
    "ver37"/If(lambda ctx: 63 > find_save_version(ctx) >= 37, Bytes(5)),
    "ver63"/If(lambda ctx: find_save_version(ctx) >= 63, Bytes(4))
)

move_to = "move_to"/Struct(
    "range"/Float32l
)

enter = "enter"/Struct(
    "first_time"/Int32ul
)

make = "make"/Struct(
    "work_timer"/Float32l
)

attack = "attack"/Struct(
    "range"/Float32l,
    "min_range"/Float32l,
    "missile_id"/Int16ul,
    "frame_delay"/Int16ul,
    "need_to_attack"/Int16ul,
    "was_same_owner"/Int16ul,
    "indirect_fire_flag"/Byte,
    "move_sprite_id"/Int16ul,
    "fight_sprite_id"/Int16ul,
    "wait_sprite_id"/Int16ul,
    "last_target_position"/vector
)

unit_action = Struct(
    "type"/Int16ul,
    "data"/If(lambda ctx: ctx.type > 0, Struct(
        "state"/IfThenElse(lambda ctx: find_version(ctx) in [Version.AOK, Version.AOC, Version.HD], Byte, Int32ul),
        "target_object_pointer"/Int32sl,
        "target_object_pointer2"/Int32sl,
        "target_object_id"/Int32sl,
        "target_object_id_2"/Int32sl,
        "position"/vector,
        "timer"/Float32l,
        "target_moved_state"/Byte,
        "task_id"/Int16ul,
        "sub_action_value"/Byte,
        "sub_actions"/LazyBound(lambda x: action_list),
        "sprite_id"/Int16sl,
        Embedded("additional"/Switch(lambda ctx: ctx._.type, {
            1: move_to,
            3: enter,
            9: attack,
            21: make
        }, default=Pass))
    ))
)

action_list = "action_list"/Struct(
    "actions"/RepeatUntil(lambda x,lst,ctx: lst[-1].type == 0, unit_action)
)

action = "action"/Struct(
    Embedded(base_moving),
    "hd_action"/If(lambda ctx: find_version(ctx) == Version.HD, Bytes(3)),
    "waiting"/Byte,
    "command_flag"/Byte,
    "selected_group_info"/If(lambda ctx: find_version(ctx) != Version.AOK, Int16ul),
    "actions"/action_list
)

base_combat = "base_combat"/Struct(
    Embedded(action),
    "formation_id"/Byte,
    "formation_row"/Byte,
    "formation_col"/Byte,
    "attack_timer"/Float32l,
    "capture_flag"/Byte,
    "multi_unified_points"/Byte,
    "large_object_radius"/Byte,
    "attack_count"/Int32ul
)

missile = "missile"/Struct(
    Embedded(base_combat),
    "max_range"/Float32l,
    "fired_from_id"/Int32ul,
    "own_base"/Byte,
    "base"/If(lambda ctx: ctx.own_base > 0, unit_type),
    If(lambda ctx: find_save_version(ctx) >= 26.16, Bytes(34))
)

waypoint = "waypoint"/Struct(
    vector,
    "facet_to_next_waypoint"/Byte,
    Padding(3)
)

gatherpoint = "gatherpoint"/Struct(
    "exists"/Int32ul,
    "position"/vector,
    "object_id"/Int32sl,
    "unit_type_id"/Int16sl
)

path = "path"/Struct(
)

order = "order"/Struct(
    "issuer"/Int32ul,
    "type"/Int32ul,
    "priority"/Int32ul,
    "target_id"/Int32ul,
    "target_player"/Int32ul,
    "target_location"/vector,
    "range"/Float32l
)

notification = "notification"/Struct(
    "caller"/Int32ul,
    "recipient"/Int32ul,
    "notification_type"/Int32ul,
    "params"/Array(3, Int32ul)
)

order_history = "order_history"/Struct(
    "order"/Int32ul,
    "action"/Int32ul,
    "time"/Int32ul,
    "position"/vector,
    "target_id"/Int32ul,
    "target_attack_category"/Int32ul,
    "target_position"/vector
)

retarget = "retarget"/Struct(
    "target_id"/Int32ul,
    "retarget_timeout"/Int32ul
)

unit_ai = "ai"/Struct(
    "mood"/Int32sl,
    "current_order"/Int32sl,
    "current_order_priority"/Int32sl,
    "current_action"/Int32sl,
    "current_target"/Int32sl,
    "current_target_type"/Int16sl,
    Padding(2),
    "current_target_location"/vector,
    "desired_target_distance"/Float32l,
    "last_action"/Int32sl,
    "last_order"/Int32sl,
    "last_target"/Int32sl,
    "last_target_type"/Int32sl,
    "last_update_type"/Int32sl,
    "idle_timer"/Int32ul,
    "idle_timeout"/Int32ul,
    "adjusted_idle_timeout"/Int32ul,
    "secondary_timer"/Int32ul,
    "look_around_timer"/Int32ul,
    "look_around_timeout"/Int32ul,
    "defend_target"/Int32sl,
    "defense_buffer"/Float32l,
    "last_world_position"/waypoint,
    "de_2006_unk"/If(lambda ctx: 26.21 > find_save_version(ctx) >= 20.06, Struct(
        "unk_float"/Float32l,
        "unk"/Int32ul
    )),
    "num_orders"/Int32ul,
    "orders"/Array(lambda ctx: ctx.num_orders, order),
    "num_notifications"/Int32ul,
    "notifications"/Array(lambda ctx: ctx.num_notifications, notification),
    "num_attacking_units"/Int32ul,
    "attacking_units"/Array(lambda ctx: ctx.num_attacking_units, Int32ul),
    "stop_after_target_killed"/Byte,
    "state"/Byte,
    "state_position_x"/Float32l,
    "state_position_y"/Float32l,
    "time_since_enemy_sighting"/Int32ul,
    "alert_mode"/Byte,
    "alert_mode_object_id"/Int32sl,
    "has_patrol_path"/Int32ul,
    If(lambda ctx: ctx.has_patrol_path > 0, path),
    "patrol_current_waypoint"/Int32ul,
    "num_order_history"/Int32ul,
    "order_history"/Array(lambda ctx: ctx.num_order_history, order_history),
    "last_retarget_time"/Int32ul,
    "randomized_retarget_timer"/Int32ul,
    "num_retarget_entries"/Int32ul,
    "retarget_entries"/Array(lambda ctx: ctx.num_retarget_entries, retarget),
    "best_unit_to_attack"/Int32sl,
    "formation_type"/Byte,
    "de_unk"/If(lambda ctx: find_version(ctx) == Version.DE, Bytes(4)),
    "de_unk_byte"/If(lambda ctx: find_save_version(ctx) >= 25.22, Byte),
    "de_unknown_2"/If(lambda ctx: find_save_version(ctx) >= 63.0 and ctx._.has_ai in (15, 17), Bytes(4))
)


combat = "combat"/Struct(
    Embedded(base_combat),
    "de_pre"/If(lambda ctx: find_version(ctx) == Version.DE and find_save_version(ctx) < 37, Bytes(4)),
    "de"/If(lambda ctx: find_version(ctx) == Version.DE, Bytes(14)),
    "de_unknown_66_3_1"/If(lambda ctx: find_save_version(ctx) >= 66.3, Bytes(4)),
    "de_2"/If(lambda ctx: find_save_version(ctx) >= 26.16, Bytes(16)),
    "de_3"/If(lambda ctx: 63 > find_save_version(ctx) >= 26.18, Bytes(1)),
    "de_4"/If(lambda ctx: find_save_version(ctx) >= 61.5, Bytes(4)),
    "de_5"/If(lambda ctx: find_save_version(ctx) >= 64.3, Bytes(19)),
    "next_volley"/Byte,
    "using_special_animation"/Byte,
    "own_base"/Byte,
    "base"/If(lambda ctx: ctx.own_base > 0, unit_type),
    "attribute_amounts"/Array(6, Int16ul),
    "decay_timer"/Int16ul,
    "raider_build_countdown"/Int32ul,
    "locked_down_count"/Int32ul,
    "inside_garrison_count"/If(lambda ctx: find_version(ctx) != Version.AOK, Byte),
    "has_ai"/Int32ul,
    "ai"/If(lambda ctx: ctx.has_ai > 0, unit_ai),
    "peek"/Peek(Bytes(5)), # TODO: figure out the right way to do this part
    "hd_position"/If(lambda ctx: find_version(ctx) == Version.HD and ctx.peek != b'\x00\xff\xff\xff\xff', Bytes(4)),
    "de_position"/If(lambda ctx: find_version(ctx) == Version.DE and ctx.peek != b'\x00\xff\xff\xff\xff', Struct(
        "position"/vector,
        "flag"/Byte,
    )),
    "town_bell_flag"/Byte,
    "town_bell_target_id"/Int32sl,
    "town_bell_target_x"/Float32l,
    "town_bell_target_y"/Float32l,
    "town_bell_target_id_2"/If(lambda ctx: find_version(ctx) != Version.AOK, Int32sl),
    "town_bell_target_type"/If(lambda ctx: find_version(ctx) != Version.AOK, Int32sl),
    "town_bell_action"/If(lambda ctx: find_version(ctx) != Version.AOK, Int32sl),
    "berserker_timer"/Float32l,
    "num_builders"/Byte,
    "num_healers"/If(lambda ctx: find_version(ctx) != Version.AOK, Byte),
    "de_unknown"/If(lambda ctx: find_save_version(ctx) >= 20.06, Int32ul),
    "de_unknown2"/If(lambda ctx: find_save_version(ctx) >= 25.01, Int32ul),
    "de_unknown3"/If(lambda ctx: 26.18 > find_save_version(ctx) >= 26.16, Bytes(5)),
    "de_unknown4"/If(lambda ctx: find_save_version(ctx) >= 26.18, Bytes(4)),
    "de_unknown5"/If(lambda ctx: find_save_version(ctx) >= 50, Bytes(48)),
    "de_unknown6"/If(lambda ctx: find_save_version(ctx) >= 61.5, Bytes(40)),
    "de_unknown7"/If(lambda ctx: find_save_version(ctx) >= 61.5, Int16ul),
    Embedded(
        IfThenElse(lambda ctx: find_save_version(ctx) >= 63.0 and ctx.has_ai == 15,
            Struct(
                "de_unknown8"/Float32l,
                "de_unknown9"/Bytes(4)
            ),
            Struct(
                "de_unknown8"/If(lambda ctx: find_save_version(ctx) >= 61.5, Bytes(2))
            )
        )
    ),
    "de_unknown_64_3_1"/If(lambda ctx: find_save_version(ctx) >= 64.3, Byte),
)

production_queue = "production_queue"/Struct(
    "unit_type_id"/Int16ul,
    "count"/Int16ul
)

building = "building"/Struct(
    Embedded(combat),
    "built"/Byte,
    "build_points"/Float32l,
    "unique_build_id"/Int32ul,
    "culture"/Byte,
    "burning"/Byte,
    "last_burn_time"/Int32ul,
    "last_garrison_time"/Int32ul,
    "relic_count"/Int32ul,
    "specific_relic_count"/Int32ul,
    "gather_point"/gatherpoint,
    "desolid_flag"/Byte,
    "pending_order"/Int32ul,
    "linked_owner"/Int32sl,
    "linked_children"/Array(lambda ctx: 3 if find_version(ctx) in (Version.DE, Version.HD) else 4, Int32sl),
    "captured_unit_count"/Byte,
    "extra_actions"/action_list,
    "research_actions"/If(lambda ctx: find_version(ctx) != Version.DE, action_list),
    "capacity"/Int16ul,
    "production_queue"/Array(lambda ctx: ctx.capacity, production_queue),
    "size"/Int16ul,
    "production_queue_total_units"/Int16ul,
    "production_queue_enabled"/Byte,
    "production_queue_actions"/action_list,
    "endpoint"/vector,
    "endpoint_2"/vector,
    "gate_locked"/Int32ul,
    "first_update"/Int32ul,
    "close_timer"/Int32ul,
    "terrain_type"/Byte,
    "semi_asleep"/Byte,
    "snow_flag"/If(lambda ctx: find_version(ctx) != Version.AOK, Byte),
    "de_flag_unk"/If(lambda ctx: find_version(ctx) == Version.DE, Byte),
    "de_unk_2"/If(lambda ctx: find_save_version(ctx) >= 20.16, Int16ul),
    "de_unk_3"/If(lambda ctx: find_save_version(ctx) >= 25.22, Byte),
    "de_unk_4"/If(lambda ctx: find_save_version(ctx) >= 26.16, Bytes(4)),
    "de_unk_5"/If(lambda ctx: find_save_version(ctx) >= 50.4, Bytes(4)),
    "de_unk_6"/If(lambda ctx: find_save_version(ctx) >= 61.5, Bytes(12)),
    "de_unk_7"/If(lambda ctx: find_save_version(ctx) >= 64.3, Bytes(4)),
    "de_unknown_66_3_2"/If(lambda ctx: find_save_version(ctx) >= 66.3, Bytes(5)),
)


# Objects that exist on the map at the start of the recorded game
existing_object = "objects"/Struct(
    "type"/Byte,
    "player_id"/Byte,
    Embedded("properties"/Switch(lambda ctx: ctx.type, {
        10: static,
        20: animated,
        25: animated,
        30: moving,
        40: action,
        50: base_combat,
        60: missile,
        70: combat,
        80: building,
        90: static
    }, default=Pass))
)

# Default values for objects, nothing of real interest
default_object = "default_object"/Struct(
    ObjectTypeEnum("type"/Byte),
    Padding(14),
    "properties"/Switch(lambda ctx: ctx.type, {
        "gaia": Padding(24),
        "object": Padding(32),
        "fish": Padding(32),
        "other": Padding(28),
    }, default="end_of_object"/Find(b'\x21\x16', 200))
)



"""Achievements."""

from construct import Array, Float32l, Int32ul, Padding, Struct

# pylint: disable=invalid-name


# Achievements snapshot taken at the start of the recorded game, not the end.
player_achievements = "ach"/Struct(
    Padding(13),
    "total_points"/Int32ul,
    Padding(26),
    "war_points"/Int32ul,
    Padding(34),
    "num_kills"/Float32l,
    Padding(28),
    "num_killed"/Float32l,
    Padding(28),
    "num_killed2"/Float32l,
    Padding(284),
    Padding(4),
    Padding(60),
    Padding(4),
    Padding(444),
    "total_food"/Float32l,
    Padding(28),
    "total_wood"/Float32l,
    Padding(28),
    "total_stone"/Float32l,
    Padding(28),
    "total_gold"/Float32l,
    Padding(444),
    "max_villagers"/Float32l,
    Padding(28),
    "max_villagers2"/Float32l,
    Padding(28),
    "relic_castle"/Float32l,
    Padding(28),
    "relic_castle2"/Float32l,
    Padding(28),
    "wonder"/Float32l,
    Padding(28),
    "relic_gold"/Float32l,
    Padding(124),
    "researched_techs"/Float32l,
    Padding(28),
    "explored_percent"/Float32l,
    Padding(4),
)

"""Achievements exist for all players but Gaia"""
achievements = Array(lambda ctx: ctx.replay.num_players - 1, player_achievements)



"""Scenario."""

import struct
from construct import (Array, Float32l, Int16ul, Int32sl, Int32ul, Padding,
                       PascalString, Peek, String, Struct, Bytes, If, IfThenElse)

from mgz.enums import DifficultyEnum, PlayerTypeEnum, AgeEnum
from mgz.util import Find, Version, find_save_version, find_version
from mgz.header.de import de_string

# pylint: disable=invalid-name, bad-continuation

# Scenario header.
scenario_header = "scenario_header"/Struct(
    "next_uid"/Int32ul,
    "scenario_version"/Float32l,
    If(lambda ctx: find_save_version(ctx) >= 61.5, Struct(
        "player_capacity"/Int32ul,
        "gaia_player_index"/Int32ul
    )),
    Array(16, "tribe_names"/String(256)),
    Array(16, "player_names"/Int32ul),
    If(lambda ctx: find_save_version(ctx) >= 66.6, Array(16, "player_data"/Struct(
        "zeros"/Int32ul,
        "active"/Int32ul,
        "string_1"/de_string,
        "string_2"/de_string,
        "unknown"/Int32ul
    ))),
    If(lambda ctx: find_save_version(ctx) >= 61.5 and find_save_version(ctx) < 66.6, Bytes(64)),
    If(lambda ctx: find_save_version(ctx) < 66.6, Array(16, "player_data"/Struct(
        "active"/Int32ul,
        "human"/Int32ul,
        "civilization"/Int32ul,
        "civ_repeat"/If(lambda ctx: find_save_version(ctx) >= 13.34, Int32ul),
        "constant"/Int32ul, # 0x04 0x00 0x00 0x00
    ))),
    Padding(5),
    "elapsed_time"/Float32l,
    "scenario_filename"/PascalString(lengthfield="scenario_filename_length"/Int16ul),
    If(lambda ctx: ctx._._.version == Version.DE, Struct(
        Padding(64),
    )),
    If(lambda ctx: find_save_version(ctx) >= 66.6, Bytes(64)),
)

# Scenarios have intro text, a bitmap, and cinematics.
messages = "messages"/Struct(
    "instruction_id"/Int32sl,
    "hints_id"/Int32sl,
    "victory_id"/Int32sl,
    "defeat_id"/Int32sl,
    "history_id"/Int32sl,
    "scouts_id"/If(lambda ctx: ctx._._.version != Version.AOK, Int32sl),
    "instructions_length"/Int16ul,
    "instructions"/Bytes(lambda ctx: ctx.instructions_length),
    "hints"/PascalString(lengthfield="hints_length"/Int16ul),
    "victory"/PascalString(lengthfield="victory_length"/Int16ul),
    "defeat"/PascalString(lengthfield="defeat_length"/Int16ul),
    "history"/PascalString(lengthfield="history_length"/Int16ul),
    "scouts"/If(lambda ctx: ctx._._.version != Version.AOK, PascalString(lengthfield="scouts_length"/Int16ul)),
    "pg_cin"/PascalString(lengthfield="pg_cin_length"/Int16ul),
    "vict_cin"/PascalString(lengthfield="vict_cin_length"/Int16ul),
    "loss_cin"/PascalString(lengthfield="loss_cin_length"/Int16ul),
    "background"/PascalString(lengthfield="background_length"/Int16ul),
    "bitmap_included"/Int32ul,
    "bitmap_x"/Int32ul,
    "bitmap_y"/Int32ul,
    Padding(2),
    # bitmap here if it is included
    Padding(64), # 16 nulls
)

# Scenario player definitions.
scenario_players = "players"/Struct(
    Array(16, "ai_names"/PascalString(lengthfield="ai_name_length"/Int16ul)),
    Array(16, "ai"/Struct(Padding(8), "file"/PascalString(lengthfield="ai_file_length"/Int32ul))),
    Padding(4),
    Array(16, "resources"/Struct(
        "gold"/Int32ul,
        "wood"/Int32ul,
        "food"/Int32ul,
        "stone"/Int32ul,
        "unk0"/Int32ul,
        "unk1"/Int32ul,
        If(lambda ctx: ctx._._._.version in (Version.DE, Version.HD),
            "unk2"/Int32ul
        )
    )),
    Array(16, Padding(1)) # 0x01 x 16
)

# Victory conditions.
victory = "victory"/Struct(
    Padding(4),
    "is_conquest"/Int32ul,
    Padding(4),
    "relics"/Int32ul,
    Padding(4),
    "explored"/Int32ul,
    Padding(4),
    "all"/Int32ul,
    "mode"/Int32ul,
    "score"/Int32ul,
    "time"/Int32ul,
)

# Disabled techs, units, and buildings.
disables = "disables"/Struct(
    Padding(4),
    Padding(64),
    If(lambda ctx: ctx._._.version != Version.DE,
        Struct(
            Array(16, "num_disabled_techs"/Int32ul),
            Array(16, Array(30, Padding(4))),
            Array(16, "num_disabled_units"/Int32ul),
            Array(16, Array(30, Padding(4))),
            Array(16, "num_disabled_buildings"/Int32ul),
            Array(16, Array(20, Padding(4))),
        )
    ),
    If(lambda ctx: ctx._._.version == Version.DE, Bytes(196)),
    If(lambda ctx: ctx._._.version == Version.HD, Bytes(644)),
    "padding"/Bytes(12)
)

# Game settings.
game_settings = "game_settings"/Struct(
    Array(16, AgeEnum("starting_ages"/Int32sl)),
    "hd"/If(lambda ctx: find_version(ctx) == Version.HD, Bytes(16)),
    Padding(4),
    Padding(8),
    "map_id"/If(lambda ctx: ctx._._.version != Version.AOK, Int32ul),
    Peek("difficulty_id"/Int32ul),
    DifficultyEnum("difficulty"/Int32ul),
    "lock_teams"/Int32ul,
    If(lambda ctx: ctx._._.version == Version.DE,
        Struct(
            Padding(29),
            If(lambda ctx: find_save_version(ctx) >= 13.07, Padding(1)),
            If(lambda ctx: find_save_version(ctx) >= 13.34, Padding(132)),
            If(lambda ctx: find_save_version(ctx) >= 20.06, Padding(1)),
            If(lambda ctx: find_save_version(ctx) >= 20.16, Padding(4)),
            If(lambda ctx: find_save_version(ctx) >= 25.02, Padding(4*16)),
            If(lambda ctx: 26.21 > find_save_version(ctx) >= 25.06, Padding(4))
        )
    ),
    Array(9, "player_info"/Struct(
        "slot"/Int32sl,
        PlayerTypeEnum("type"/Int32ul),
        "name"/PascalString(lengthfield="name_length"/Int32ul)
    )),
    Padding(36),
    Padding(4),
    IfThenElse(lambda ctx: ctx._._.version == Version.DE,
        Struct(
            If(lambda ctx: find_save_version(ctx) < 13.34, Find(struct.pack('<d', 2.2), None)),
            If(lambda ctx: 25.06 > find_save_version(ctx) >= 13.34, Find(struct.pack('<d', 2.4), None)),
            If(lambda ctx: 25.22 > find_save_version(ctx) >= 25.06, Find(struct.pack('<d', 2.5), None)),
            If(lambda ctx: 26.16 > find_save_version(ctx) >= 25.22, Find(struct.pack('<d', 2.6), None)),
            If(lambda ctx: 26.21 > find_save_version(ctx) >= 26.16, Find(struct.pack('<d', 3.0), None)),
            If(lambda ctx: 37 > find_save_version(ctx) >= 26.21, Find(struct.pack('<d', 3.2), None)),
            If(lambda ctx: 61.5 > find_save_version(ctx) >= 37, Find(struct.pack('<d', 3.5), None)),
            If(lambda ctx: 63 > find_save_version(ctx) >= 61.5, Find(struct.pack('<d', 3.6), None)),
            If(lambda ctx: 64.3 > find_save_version(ctx) >= 63, Find(struct.pack('<d', 3.9), None)),
            If(lambda ctx: 66.3 > find_save_version(ctx) >= 64.3, Find(struct.pack('<d', 4.1), None)),
            If(lambda ctx: find_save_version(ctx) >= 66.3, Find(struct.pack('<d', 4.5), None)),
        ),
        "end_of_game_settings"/Find(b'\x9a\x99\x99\x99\x99\x99\xf9\\x3f', None),
    ),
)


# Triggers.
triggers = "triggers"/Struct(
    Padding(1),
    "num_triggers"/Int32ul,
    # parse if num > 0
    "de"/If(lambda ctx: ctx._._.version == Version.DE,
         Padding(1032)
    )
)



# Scenario metadata.
scenario = "scenario"/Struct(
    scenario_header,
    messages,
    scenario_players,
    victory,
    Padding(12544), # unknown
    disables,
    game_settings,
    triggers
)



"""UserPatch 1.4+ achievements - appended to recorded game."""

from construct import (Array, Byte, Flag, Int16ul, Int32sl, Int32ul, Padding,
                       Peek, Struct, Bytes)

from mgz.util import TimeSecAdapter

# pylint: disable=invalid-name


military = "military"/Struct(
    "score"/Int16ul,
    "units_killed"/Int16ul,
    "hit_points_killed"/Int16ul,
    "units_lost"/Int16ul,
    "buildings_razed"/Int16ul,
    "hit_points_razed"/Int16ul,
    "buildings_lost"/Int16ul,
    "units_converted"/Int16ul
)

economy = "economy"/Struct(
    "score"/Int16ul,
    Padding(2),
    "food_collected"/Int32ul,
    "wood_collected"/Int32ul,
    "stone_collected"/Int32ul,
    "gold_collected"/Int32ul,
    "tribute_sent"/Int16ul,
    "tribute_received"/Int16ul,
    "trade_gold"/Int16ul,
    "relic_gold"/Int16ul
)

technology = "technology"/Struct(
    "score"/Int16ul,
    Padding(2),
    Peek("feudal_time_int"/Int32sl),
    TimeSecAdapter("feudal_time"/Int32sl),
    Peek("castle_time_int"/Int32sl),
    TimeSecAdapter("castle_time"/Int32sl),
    Peek("imperial_time_int"/Int32sl),
    TimeSecAdapter("imperial_time"/Int32sl),
    "explored_percent"/Byte,
    "research_count"/Byte,
    "research_percent"/Byte
)

society = "society"/Struct(
    "score"/Int16ul,
    "total_wonders"/Byte,
    "total_castles"/Byte,
    "relics_captured"/Byte,
    Padding(1),
    "villager_high"/Int16ul,
)

achievements = "achievements"/Struct(
    "player_name"/Bytes(16),
    "total_score"/Int16ul,
    Array(8, "total_scores"/Int16ul),
    "victory"/Flag,
    "civilization"/Byte,
    "color_id"/Byte,
    "team"/Byte,
    "ally_count"/Byte,
    "random_civ"/Flag, # never actually set
    "mvp"/Flag,
    Padding(3),
    "result"/Int32ul,
    military,
    Padding(32),
    economy,
    Padding(16),
    technology,
    Padding(1),
    society,
    Padding(84),
)



"""Fast(er) parsing for recorded game headers."""
import io
import hashlib
import re
import struct
import uuid
import zlib

from mgz.util import get_version, unpack, Version, as_hex

ZLIB_WBITS = -15
CLASSES = [b'\x0a', b'\x1e', b'\x46', b'\x50', b'\x14']
BLOCK_END = b'\x00\x0b'
REGEXES = {}
SKIP_OBJECTS = [
    (b'\x1e\x00\x87\x02', 252)  # 647: junk DE object, thousands per file
]


def _compile_object_search():
    """Compile regular expressions for object searching."""
    class_or = b'(' + b'|'.join(CLASSES) + b')'
    for i in range(9):
        expr = class_or + struct.pack('b', i) + b'(?!\xff\xff)(?!\x00\x00)[\x00-\xff]{4}\xff\xff\xff\xff[^\xff]'
        REGEXES[i] = re.compile(expr)


_compile_object_search()


def aoc_string(data):
    """Read AOC string."""
    length = unpack('<h', data)
    return data.read(length)


def int_prefixed_string(data):
    """Read length prefixed (4 byte) string."""
    length = unpack('<I', data)
    return data.read(length)


def de_string(data):
    """Read DE string."""
    assert data.read(2) == b'\x60\x0a'
    length = unpack('<h', data)
    return unpack(f'<{length}s', data)


def hd_string(data):
    """Read HD string."""
    length = unpack('<h', data)
    assert data.read(2) == b'\x60\x0a'
    return unpack(f'<{length}s', data)


def parse_object(data, offset):
    """Parse an object."""
    class_id, object_id, instance_id, pos_x, pos_y = struct.unpack_from('<bxH14xIxff', data, offset)
    return dict(
        class_id=class_id,
        object_id=object_id,
        instance_id=instance_id,
        position=dict(
            x=pos_x,
            y=pos_y
        )
    )


def object_block(data, pos, player_number, index):
    """Parse a block of objects."""
    objects = []
    offset = None
    while True:
        if not offset:
            match = REGEXES[player_number].search(data, pos, pos + 10000)
            end = data.find(BLOCK_END, pos) - pos + len(BLOCK_END)
            if match is None:
                break
            offset = match.start() - pos
            while end + 8 < offset:
                end += data.find(BLOCK_END, pos + end) - (pos + end) + len(BLOCK_END)
        if end + 8 == offset:
            break
        pos += offset
        # Speed optimization: Skip specified fixed-length objects.
        test = data[pos:pos + 4]
        for fingerprint, offset in SKIP_OBJECTS:
            if test == fingerprint:
                break
        else:
            objects.append(dict(parse_object(data, pos), index=index))
        offset = None
        pos += 31
    return objects, pos + end


def parse_mod(header, num_players, version):
    """Parse Userpatch mod version."""
    cur = header.tell()
    name_length = unpack(f'<xx{num_players}x36x5xh', header)
    resources = unpack(f'<{name_length + 1}xIx', header)
    values = unpack(f'<{resources}f', header)
    header.seek(cur)
    if version is Version.USERPATCH15:
        number = int(values[198])
        return number // 1000, '.'.join(list(str(number % 1000)))


def parse_player(header, player_number, num_players, save):
    """Parse a player (and objects)."""
    rep = 9
    if save >= 61.5:
        rep = num_players
    type_, *diplomacy, name_length = unpack(f'<bx{num_players}x{rep}i5xh', header)
    name, resources = unpack(f'<{name_length - 1}s2xIx', header)
    resources_len = 8 if save >= 63 else 4
    header.read(resources * resources_len)
    start_x, start_y, civilization_id, color_id = unpack('<xff9xb3xbx', header)
    offset = header.tell()
    data = header.read()
    # Skips thousands of bytes that are not easy to parse.
    object_start = re.search(b'\x0b\x00.\x00\x00\x00\x02\x00\x00', data, re.DOTALL)
    if not object_start:
        raise RuntimeError("could not find object start")
    start = object_start.end()
    objects, end = object_block(data, start, player_number, 0)
    sleeping, end = object_block(data, end, player_number, 1)
    doppel, end = object_block(data, end, player_number, 2)
    if data[end + 8:end + 10] == BLOCK_END:
        end += 10
    if data[end:end + 2] == BLOCK_END:
        end += 2
    header.seek(offset + end)
    device = 0
    if save >= 37:
        offset = header.tell()
        data = header.read(100)
        device = data[8]
        # Jump to the end of player data
        player_end = re.search(b'\xff\xff\xff\xff\xff\xff\xff\xff.\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x0b', data, re.DOTALL)
        if not player_end:
            # only a failure if this is not the last player, since we seek to the next block anyway
            # this issue happens on restored games
            if player_number < num_players - 1:
                raise RuntimeError("could not find player end")
        else:
            header.seek(offset + player_end.end())

    return dict(
        number=player_number,
        type=type_,
        name=name,
        diplomacy=diplomacy,
        civilization_id=civilization_id,
        color_id=color_id,
        objects=objects + sleeping + doppel,
        position=dict(
            x=start_x,
            y=start_y
        )
    ), device


def parse_lobby(data, version, save):
    """Parse lobby data."""
    if version is Version.DE:
        data.read(5)
        if save >= 20.06:
            data.read(9)
        if save >= 26.16:
            data.read(5)
        if save >= 37:
            data.read(8)
        if save >= 64.3:
            data.read(16)
        if save >= 66.3:
            data.read(1)
    data.read(8)
    if version not in (Version.DE, Version.HD):
        data.read(1)
    reveal_map_id, map_size, population, game_type_id, lock_teams = unpack('I4xIIbb', data)
    if version in (Version.DE, Version.HD):
        data.read(5)
        if save >= 13.13:
            data.read(4)
        if save >= 25.22:
            data.read(1)
    chat = []
    for _ in range(0, unpack('<I', data)):
        message = data.read(unpack('<I', data)).strip(b'\x00')
        if len(message) > 0:
            chat.append(message)
    seed = None
    if version is Version.DE:
        seed = unpack('<i', data)
    return dict(
        reveal_map_id=reveal_map_id,
        map_size=map_size,
        population=population * (25 if version not in (Version.DE, Version.HD) else 1),
        game_type_id=game_type_id,
        lock_teams=lock_teams == 1,
        chat=chat,
        seed=seed
    )


def parse_map(data, version, save):
    """Parse map."""
    tile_format = '<xbbx'
    if version is Version.DE:
        tile_format = '<bxb6x'
        if save >= 62.0:
            tile_format = '<bxxb6x'
        data.read(8)
    size_x, size_y, zone_num = unpack('<III', data)
    tile_num = size_x * size_y
    for _ in range(zone_num):
        if version in (Version.DE, Version.HD):
            data.read(2048 + (tile_num * 2))
        else:
            data.read(1275 + tile_num)
        num_floats = unpack('<I', data)
        data.read(num_floats * 4)
        data.read(4)
    all_visible = unpack('<bx', data)
    tiles = [unpack(tile_format, data) for _ in range(tile_num)]
    num_data = unpack('<I4x', data)
    data.read(num_data * 4)
    for _ in range(0, num_data):
        num_obs = unpack('<I', data)
        data.read(num_obs * 8)
    x2, y2 = unpack('<II', data)
    data.read(x2 * y2 * 4)
    if save >= 61.5:
        data.read(x2 * y2 * 4)
    restore_time = unpack('<I', data)
    #if restore_time > 0:
    #    raise RuntimeError("restored matches can't be parsed yet")
    return dict(
        all_visible=all_visible == 1,
        restore_time=restore_time,
        dimension=size_x,
        tiles=tiles
    )


def parse_scenario(data, num_players, version, save):
    """Parse scenario section."""
    scenario_version = unpack('<f', data)
    data.read(4)
    if save >= 61.5:
        data.read(4)
        if save < 66.6:
            data.read(4)
    data.read(16 * 256)
    data.read(16 * 4)
    if save >= 66.6:
        for i in range(0, 16):
            data.read(8)
            de_string(data)
            de_string(data)
            data.read(4)
    if save >= 61.5 and save < 66.6:
        data.read(64)
    if save < 66.6:
        for i in range(0, 16):
            data.read(12)
            if save >= 13.34:
                data.read(4)
            data.read(4)
    data.read(5)
    elapsed_time = unpack('<f', data)
    scenario_filename = aoc_string(data)
    if version is Version.DE:
        data.read(64)
    if save >= 66.6:
        data.read(68)
    data.read(20)
    instructions = aoc_string(data)
    for _ in range(0, 9):
        aoc_string(data)
    data.read(78)
    for _ in range(0, 16):
        aoc_string(data)
    data.read(196)
    for _ in range(0, 16):
        data.read(24)
        if version in (Version.DE, Version.HD):
            data.read(4)
    data.read(12672)
    if version is Version.DE:
        data.read(196)
    else:
        for _ in range(0, 16):
            data.read(332)
    if version is Version.HD:
        data.read(644)
    data.read(88)
    if version is Version.HD:
        data.read(16)
    map_id, difficulty_id = unpack('<II', data)
    remainder = data.read()
    if version is Version.DE:
        if save >= 66.3:
            settings_version = 4.5
        elif save >= 64.3:
            settings_version = 4.1
        elif save >= 63:
            settings_version = 3.9
        elif save >= 61.5:
            settings_version = 3.6
        elif save >= 37:
            settings_version = 3.5
        elif save >= 26.21:
            settings_version = 3.2
        elif save >= 26.16:
            settings_version = 3.0
        elif save >= 25.22:
            settings_version = 2.6
        elif save >= 25.06:
            settings_version = 2.5
        elif save >= 13.34:
            settings_version = 2.4
        else:
            settings_version = 2.2
        end = remainder.find(struct.pack('<d', settings_version)) + 8
    else:
        end = remainder.find(b'\x9a\x99\x99\x99\x99\x99\xf9\x3f') + 13
    data.seek(end - len(remainder), 1)

    if version is Version.DE:
        data.read(1)
        n_triggers = unpack("<I", data)

        for _ in range(n_triggers):
            data.read(22)
            data.read(4)

            description = int_prefixed_string(data)
            name = int_prefixed_string(data)
            short_description = int_prefixed_string(data)

            n_effects = unpack("<I", data)

            for _ in range(n_effects):
                data.read(216)

                text = int_prefixed_string(data)
                sound = int_prefixed_string(data)

            data.read(n_effects * 4)
            n_condition = unpack("<I", data)

            data.read(n_condition * 125)

        trigger_list_order = unpack(f"<{n_triggers}I", data)

        data.read(1032)  # default!

    return dict(
        map_id=map_id,
        difficulty_id=difficulty_id,
        instructions=instructions,
        scenario_filename=scenario_filename,
    )


def string_block(data):
    """Parse DE header string block."""
    strings = []
    while True:
        crc = unpack("<I", data)
        if 255 > crc > 0:
            break
        strings.append(de_string(data).decode('utf-8').split(':'))
    return strings


def parse_de(data, version, save, skip=False):
    """Parse DE-specific header."""
    if version is not Version.DE:
        return None
    build = None
    if save >= 25.22 and not skip:
        build = unpack('<I', data)
    timestamp = None
    if save >= 26.16 and not skip:
        timestamp = unpack('<I', data)  # missing on console (?)
    data.read(12)
    dlc_ids = []
    for i in range(0, unpack('<I', data)):
        dlc_ids.append(unpack('<I', data))
    data.read(4)
    if save >= 61.5:
        map_dimension = unpack('<I', data)
    else:
        difficulty_id = unpack('<I', data)
    data.read(4)
    rms_map_id = unpack('<I', data)
    data.read(4)
    victory_type_id = unpack('<I', data)
    starting_resources_id = unpack('<I', data)
    starting_age_id = unpack('<I', data)
    ending_age_id = unpack('<I', data)
    data.read(12)
    speed = unpack('<f', data)
    treaty_length = unpack('<I', data)
    population_limit = unpack('<I', data)
    num_players = unpack('<I', data)
    data.read(14)
    if save >= 61.5:
        # not sure if this is difficulty under 61.5 or not
        difficulty_id = unpack('<B', data)
    random_positions, all_technologies = unpack('<bb', data)
    data.read(1)
    lock_teams = unpack('<b', data)
    lock_speed = unpack('<b', data)
    multiplayer = unpack('<b', data)
    cheats = unpack('<b', data)
    record_game = unpack('<b', data)
    animals_enabled = unpack('<b', data)
    predators_enabled = unpack('<b', data)
    turbo_enabled = unpack('<b', data)
    shared_exploration = unpack('<b', data)
    team_positions = unpack('<b', data)
    data.read(12)
    if save >= 25.06:
        data.read(1)
    if save > 50:
        data.read(1)
    players = []
    for _ in range(num_players if 66.3 > save >= 37 else 8):
        data.read(4)
        color_id = unpack('<i', data)
        data.read(2)
        team_id = unpack('<b', data)
        data.read(9)
        civilization_id = unpack('<I', data)
        custom_civ_selection = None
        if save >= 61.5:
            custom_civ_count = unpack('<I', data)
            if save >= 63.0 and custom_civ_count > 0:
                custom_civ_selection = []
                for _ in range(custom_civ_count):
                    custom_civ_selection.append(unpack('<I', data))
        de_string(data)
        data.read(1)
        ai_name = de_string(data)
        if save >= 66.3:
            censored_name = de_string(data)
        name = de_string(data)
        type = unpack('<I', data)
        profile_id, number = unpack('<I4xi', data)
        if save < 25.22:
            data.read(8)
        prefer_random = unpack('b', data)
        data.read(1)
        if save >= 25.06:
            data.read(8)
        if save >= 64.3:
            data.read(4)

        players.append(dict(
            number=number,
            color_id=color_id,
            team_id=team_id,
            ai_name=ai_name,
            name=name,
            censored_name=censored_name if save >= 66.3 else name,
            type=type,
            profile_id=profile_id,
            civilization_id=civilization_id,
            custom_civ_selection=custom_civ_selection,
            prefer_random=prefer_random == 1
        ))
    data.read(12)
    if 66.3 > save >= 37:
        for _ in range(8 - num_players):
            if save >= 61.5:
                data.read(4)
            data.read(12)
            de_string(data)
            data.read(1)
            de_string(data)
            de_string(data)
            data.read(38)
            if save >= 64.3:
                data.read(4)
    data.read(4)
    rated = unpack('b', data)
    allow_specs = unpack('b', data)
    visibility = unpack('<I', data)
    hidden_civs = unpack('b', data)
    data.read(1)
    spec_delay = unpack('<I', data)
    data.read(1)
    strings = string_block(data)
    data.read(8)
    for _ in range(20):
        strings += string_block(data)
    data.read(4)
    if save < 25.22:
        data.read(236)
    if save >= 25.22:
        data.seek(-4, 1)
        l = unpack('<I', data)
        data.read(l * 4)
    for _ in range(unpack('<Q', data)):
        data.read(4)
        de_string(data)
        data.read(4)
    if save >= 25.02:
        data.read(8)
    guid = data.read(16)
    lobby = de_string(data)
    if save >= 25.22:
        data.read(8)
    mod = de_string(data)
    data.read(33)
    if save >= 20.06:
        data.read(1)
    if save >= 20.16:
        data.read(8)
    if save >= 25.06:
        data.read(21)
    if save >= 25.22:
        data.read(4)
    if save >= 26.16:
        data.read(8)
    if save >= 37:
        data.read(3)
    if save > 50:
        data.read(8)
    if save >= 61.5:
        data.read(1)
    if save >= 63:
        data.read(5)
    if save >= 66.3:
        c = unpack('<I', data)
        data.read(12)
        data.read(c * 4)
    if not skip:
        de_string(data)
        data.read(8)
        if save >= 37:
            timestamp, x = unpack('<II', data)
    rms_mod_id = None
    rms_filename = None
    for s in strings:
        if s[0] == 'SUBSCRIBEDMODS' and s[1] == 'RANDOM_MAPS':
            rms_mod_id = s[3].split('_')[0]
            rms_filename = s[2]
    return dict(
        players=players,
        guid=str(uuid.UUID(bytes=guid)),
        hash=hashlib.sha1(guid),
        lobby=lobby.decode('utf-8'),
        mod=mod.decode('utf-8'),
        difficulty_id=difficulty_id,
        victory_type_id=victory_type_id,
        starting_resources_id=starting_resources_id,
        starting_age_id=starting_age_id - 2 if starting_age_id > 0 else 0,
        ending_age_id=ending_age_id - 2 if ending_age_id > 0 else 0,
        speed=speed,
        population_limit=population_limit,
        treaty_length=treaty_length,
        team_together=not bool(random_positions),
        all_technologies=bool(all_technologies),
        lock_teams=bool(lock_teams),
        lock_speed=bool(lock_speed),
        multiplayer=bool(multiplayer),
        cheats=bool(cheats),
        record_game=bool(record_game),
        animals_enabled=bool(animals_enabled),
        predators_enabled=bool(predators_enabled),
        turbo_enabled=bool(turbo_enabled),
        shared_exploration=bool(shared_exploration),
        team_positions=bool(team_positions),
        build=build,
        timestamp=timestamp,
        spec_delay=spec_delay,
        rated=rated == 1,
        allow_specs=bool(allow_specs),
        hidden_civs=bool(hidden_civs),
        visibility_id=visibility,
        rms_mod_id=rms_mod_id,
        rms_map_id=rms_map_id,
        rms_filename=rms_filename,
        dlc_ids=dlc_ids
    )


def parse_hd(data, version, save):
    """Parse HD-specifc header."""
    if version is not Version.HD or save <= 12.34:
        return None
    data.read(12)
    dlc_count = unpack('<I', data)
    data.read(dlc_count * 4)
    data.read(4)
    difficulty_id, map_id = unpack('<II', data)
    data.read(80)
    players = []
    for _ in range(8):
        data.read(4)
        color_id = unpack('<i', data)
        data.read(12)
        civilization_id = unpack('<I', data)
        hd_string(data)
        data.read(1)
        hd_string(data)
        name = hd_string(data)
        data.read(4)
        steam_id, number = unpack('<Qi', data)
        data.read(8)
        if name:
            players.append(dict(
                number=number,
                color_id=color_id,
                name=name,
                profile_id=steam_id,
                civilization_id=civilization_id
            ))
    data.read(26)
    hd_string(data)
    data.read(8)
    hd_string(data)
    data.read(8)
    hd_string(data)
    data.read(8)
    guid = data.read(16)
    lobby = hd_string(data)
    mod = hd_string(data)
    data.read(8)
    hd_string(data)
    data.read(4)
    return dict(
        players=players,
        guid=str(uuid.UUID(bytes=guid)),
        lobby=lobby.decode('utf-8'),
        mod=mod.decode('utf-8'),
        map_id=map_id,
        difficulty_id=difficulty_id
    )


def decompress(data):
    """Decompress header bytes."""
    prefix_size = 8
    header_len, _ = unpack('<II', data)
    zlib_header = data.read(header_len - prefix_size)
    return io.BytesIO(zlib.decompress(zlib_header, wbits=ZLIB_WBITS))


def parse_version(header, data):
    """Parse and compute game version."""
    log = unpack('<I', data)
    game, save = unpack('<7sxf', header)
    if save == -1:
        save = unpack('<I', header)
        if save == 37:
            save = 37.0
        else:
            save /= (1<<16)
    version = get_version(game.decode('ascii'), round(save, 2), log)
    return version, game.decode('ascii'), round(save, 2), log


def parse_players(header, num_players, version, save):
    """Parse all players."""
    cur = header.tell()
    gaia = b'Gaia' if version in (Version.DE, Version.HD) else b'GAIA'
    anchor = header.read().find(b'\x05\x00' + gaia + b'\x00')
    rev = 43
    if save >= 61.5:
        rev = 7 + (num_players * 4)
    header.seek(cur + anchor - num_players - rev)
    mod = parse_mod(header, num_players, version)
    players = [parse_player(header, number, num_players, save) for number in range(num_players)]
    cur = header.tell()
    pv = b'\x00\x00\x00@'
    if save >= 61.5:
        pv = b'\x66\x66\x06\x40'
    points_version = header.read().find(pv)
    header.seek(cur)
    header.read(points_version)
    for _ in range(num_players):
        version = unpack('<f', header)
        entries = unpack('<i', header)
        header.read(5 + (entries * 44))
        points = unpack('<i', header)
        header.read(8 + (points * 32))
    return [p[0] for p in players], mod, players[0][1]


def parse_metadata(header, save, skip_ai=True):
    """Parse recorded game metadata."""
    ai = unpack('<I', header)

    if ai > 0:
        if not skip_ai:
            raise RuntimeError("don't know how to parse ai")

        offset = header.tell()
        data = header.read()
        # Jump to the end of ai data
        ai_end = re.search(
            b'\00' * 4096,
            data)
        if not ai_end:
            raise RuntimeError("could not find ai end")
        header.seek(offset + ai_end.end())

    game_speed, owner_id, num_players, cheats = unpack('<24xf17xhbxb', header)

    if save < 61.5:
        header.read(60)
    else:
        header.read(24 + (num_players * 4))

    return dict(
        speed=game_speed,
        owner_id=owner_id,
        cheats=cheats == 1
    ), num_players


def parse(data):
    """Parse recorded game header."""
    try:
        header = decompress(data)
        version, game, save, log = parse_version(header, data)
        if version not in (Version.USERPATCH15, Version.DE, Version.HD):
            raise RuntimeError(f"{version} not supported")
        de = parse_de(header, version, save)
        hd = parse_hd(header, version, save)
        metadata, num_players = parse_metadata(header, save)
        map_ = parse_map(header, version, save)
        players, mod, device = parse_players(header, num_players, version, save)
        scenario = parse_scenario(header, num_players, version, save)
        lobby = parse_lobby(header, version, save)
    except (struct.error, zlib.error, AssertionError, MemoryError, ValueError) as e:
        raise RuntimeError(f"could not parse: {e}")
    return dict(
        version=version,
        game_version=game,
        save_version=save,
        log_version=log,
        players=players,
        map=map_,
        de=de,
        hd=hd,
        mod=de.get('dlc_ids') if de else mod,
        metadata=metadata,
        scenario=scenario,
        lobby=lobby,
        device=device
    )



"""Lobby."""

from construct import Array, Byte, Bytes, Flag, Int32ul, Padding, Peek, Struct, If, Computed, Embedded, Int32sl

from mgz.enums import GameTypeEnum, RevealMapEnum
from mgz.util import Version, find_save_version

# pylint: disable=invalid-name, bad-continuation


# Player inputs in the lobby, and several host settings.
lobby = "lobby"/Struct(
    If(lambda ctx: find_save_version(ctx) >= 13.34, Padding(5)),
    If(lambda ctx: find_save_version(ctx) >= 20.06, Padding(9)),
    If(lambda ctx: find_save_version(ctx) >= 26.16, Bytes(5)),
    If(lambda ctx: find_save_version(ctx) >= 37, Bytes(8)),
    If(lambda ctx: find_save_version(ctx) >= 64.3, Bytes(16)),
    If(lambda ctx: find_save_version(ctx) >= 66.3, Bytes(1)),
    Array(8, "teams"/Byte), # team number selected by each player
    If(lambda ctx: ctx._.version not in (Version.DE, Version.HD),
        Padding(1),
    ),
    Peek("reveal_map_id"/Int32ul),
    RevealMapEnum("reveal_map"/Int32ul),
    "fog_of_war"/Int32ul,
    "map_size"/Int32ul,
    "population_limit_encoded"/Int32ul,
    "population_limit"/Computed(lambda ctx: ctx.population_limit_encoded * (25 if ctx._.version in [Version.USERPATCH14, Version.USERPATCH15] else 1)),
    Embedded(If(lambda ctx: ctx._.version != Version.AOK,
        Struct(
            Peek("game_type_id"/Byte),
            GameTypeEnum("game_type"/Byte),
            "lock_teams"/Flag
        )
    )),
    If(lambda ctx: ctx._.version in (Version.DE, Version.HD),
        Struct(
            "treaty_length"/Byte,
            "cheat_codes_used"/Int32ul,
            If(lambda ctx: find_save_version(ctx) >= 13.13, Padding(4)),
            If(lambda ctx: find_save_version(ctx) >= 25.22, Padding(1)),
        )
    ),
    Embedded(If(lambda ctx: ctx._.version != Version.AOK,
        Struct(
            "num_chat"/Int32ul,
            Array(
                lambda ctx: ctx.num_chat, # pre-game chat messages
                "messages"/Struct(
                    "message_length"/Int32ul,
                    "message"/Bytes(lambda ctx: ctx.message_length - (1 if ctx.message_length > 0 else 0)),
                    If(lambda ctx: ctx.message_length > 0, Padding(1))
                )
            )
        )
    )),
    "de"/If(lambda ctx: ctx._.version == Version.DE,
        Struct(
            "map_seed"/If(lambda ctx: find_save_version(ctx) >= 13.08, Int32sl),
            Bytes(10),
            If(lambda ctx: find_save_version(ctx) >= 26.16, Bytes(4)),
            If(lambda ctx: find_save_version(ctx) >= 37, Bytes(4)),
            If(lambda ctx: find_save_version(ctx) >= 50, Bytes(1)),
            If(lambda ctx: find_save_version(ctx) >= 66.3, Bytes(4)),
        )
    )
)



"""Definitive Edition structure."""

from construct import (
    Struct, Int32ul, Float32l, Array, Padding, Flag, If,
    Byte, Int16ul, Bytes, Int32sl, Peek, Const, RepeatUntil,
    Int64ul, Computed, IfThenElse
)

from mgz.enums import VictoryEnum, ResourceLevelEnum, AgeEnum, PlayerTypeEnum, DifficultyEnum
from mgz.util import find_save_version

# pylint: disable=invalid-name, bad-continuation

de_string = Struct(
    Const(b"\x60\x0A"),
    "length"/Int16ul,
    "value"/Bytes(lambda ctx: ctx.length)
)

separator = Const(b"\xa3_\x02\x00")

player = Struct(
    "dlc_id"/Int32ul,
    "color_id"/Int32sl,
    "selected_color"/Byte,
    "selected_team_id"/Byte,
    "resolved_team_id"/Byte,
    "dat_crc"/Bytes(8),
    "mp_game_version"/Byte,
    "civ_id"/Int32ul,
    "custom_civ_count"/IfThenElse(lambda ctx: find_save_version(ctx) >= 61.5, Int32ul, Computed(lambda _: 0)),
    "custom_civ_ids"/Array(lambda ctx: ctx.custom_civ_count, Int32ul),
    "ai_type"/de_string,
    "ai_civ_name_index"/Byte,
    "ai_name"/de_string,
    "censored_name"/If(lambda ctx: find_save_version(ctx) >= 66.3,de_string),
    "name"/de_string,
    "type"/PlayerTypeEnum(Int32ul),
    "profile_id"/Int32sl, # -1 if it's an empty slot
    "ai_unknown" / Int32sl, # Often 0, sometimes -1 for AI players
    "player_number"/Int32sl,
    "hd_rm_elo"/If(lambda ctx: find_save_version(ctx) < 25.22, Int32ul),
    "hd_dm_elo"/If(lambda ctx: find_save_version(ctx) < 25.22, Int32ul),
    "prefer_random"/Flag,
    "custom_ai"/Flag,
    If(lambda ctx: find_save_version(ctx) >= 25.06, "handicap"/Bytes(8)),
    If(lambda ctx: find_save_version(ctx) >= 64.3, "unknown_de_64_3" / Int32ul),
)

string_block = Struct(
    "strings"/RepeatUntil(lambda x, lst, ctx: 255 > lst[-1].crc > 0, Struct(
        "crc"/Int32ul,
        "string"/If(lambda ctx: ctx.crc == 0 or ctx.crc > 255, de_string)
    ))
)

de = "de"/Struct(
    "build"/If(lambda ctx: find_save_version(ctx) >= 25.22, Int32ul),
    "timestamp"/If(lambda ctx: find_save_version(ctx) >= 26.16, Int32ul),
    "version"/Float32l,
    "interval_version"/Int32ul,
    "game_options_version"/Int32ul,
    "dlc_count"/Int32ul,
    "dlc_ids"/Array(lambda ctx: ctx.dlc_count, Int32ul),
    "dataset_ref"/Int32ul,
    Peek("difficulty_id"/Int32ul), # map size now?
    "diff"/Int32ul,
    "selected_map_id"/Int32ul,
    "resolved_map_id"/Int32ul,
    "reveal_map"/Int32ul,
    Peek("victory_type_id"/Int32ul),
    VictoryEnum("victory_type"/Int32ul),
    Peek("starting_resources_id"/Int32ul),
    ResourceLevelEnum("starting_resources"/Int32ul),
    "starting_age_id"/Int32ul,
    "starting_age"/AgeEnum(Computed(lambda ctx: ctx.starting_age_id - 2)),
    "ending_age_id"/Int32ul,
    "ending_age"/AgeEnum(Computed(lambda ctx: ctx.ending_age_id - 2)),
    "game_type"/Int32ul,
    separator,
    separator,
    "speed"/Float32l,
    "treaty_length"/Int32ul,
    "population_limit"/Int32ul,
    "num_players"/Int32ul,
    "unused_player_color"/Int32ul,
    "victory_amount"/Int32sl,
    "unk_byte"/If(lambda ctx: find_save_version(ctx) >= 61.5, Flag),
    separator,
    "trade_enabled"/Flag,
    "team_bonus_disabled"/Flag,
    "random_positions"/Flag,
    "all_techs"/Flag,
    "num_starting_units"/Byte,
    "lock_teams"/Flag,
    "lock_speed"/Flag,
    "multiplayer"/Flag,
    "cheats"/Flag,
    "record_game"/Flag,
    "animals_enabled"/Flag,
    "predators_enabled"/Flag,
    "turbo_enabled"/Flag,
    "shared_exploration"/Flag,
    "team_positions"/Flag,
    "sub_game_mode"/If(lambda ctx: find_save_version(ctx) >= 13.34, Int32ul),
    "battle_royale_time"/If(lambda ctx: find_save_version(ctx) >= 13.34, Int32ul),
    "handicap"/If(lambda ctx: find_save_version(ctx) >= 25.06, Flag),
    "unk"/If(lambda ctx: find_save_version(ctx) >= 50, Flag),
    separator,
    "players"/Array(lambda ctx: ctx.num_players if find_save_version(ctx) >= 37 and find_save_version(ctx) < 66.3 else 8, player),
    Bytes(9),
    "fog_of_war"/Flag,
    "cheat_notifications"/Flag,
    "colored_chat"/Flag,
    "empty_slots"/If(lambda ctx: find_save_version(ctx) >= 37 and find_save_version(ctx) < 66.3, Array(lambda ctx: 8 - ctx.num_players, Struct(
        "unk"/If(lambda ctx: find_save_version(ctx) >= 61.5, Int32ul),
        "i0x"/Int32ul,
        "i0a"/Int32ul,
        "i0b"/Int32ul,
        "s1"/de_string,
        "a2"/Bytes(1),
        "s2"/de_string,
        "s3"/de_string,
        "a3"/Bytes(22),
        "i1"/Int32ul,
        "i2"/Int32ul,
        "a4"/Bytes(8),
        If(lambda ctx: find_save_version(ctx) >= 64.3, "unknown_de_64_3" / Int32ul),
    ))),
    separator,
    "ranked"/Flag,
    "allow_specs"/Flag,
    "lobby_visibility"/Int32ul,
    "hidden_civs"/Flag,
    "matchmaking"/Flag,
    "spec_delay"/If(lambda ctx: find_save_version(ctx) >= 13.13, Int32ul),
    "scenario_civ"/If(lambda ctx: find_save_version(ctx) >= 13.13, Byte),
    "rms_strings"/string_block,
    Bytes(8),
    "other_strings"/Array(20, string_block),
    "num_sn"/Int32ul,
    "strategic_numbers"/Array(lambda ctx: ctx.num_sn if find_save_version(ctx) >= 25.22 else 59, Int32sl),
    "num_ai_files"/Int64ul,
    "ai_files"/Array(lambda ctx: ctx.num_ai_files, Struct(
        Bytes(4),
        "name"/de_string,
        Bytes(4),
    )),
    If(lambda ctx: find_save_version(ctx) >= 25.02, Bytes(8)),
    "guid"/Bytes(16),
    "lobby_name"/de_string,
    If(lambda ctx: find_save_version(ctx) >= 25.22, Bytes(8)),
    "modded_dataset"/de_string,
    Bytes(19),
    If(lambda ctx: find_save_version(ctx) >= 13.13, Bytes(5)),
    If(lambda ctx: find_save_version(ctx) >= 13.17, Bytes(9)),
    If(lambda ctx: find_save_version(ctx) >= 20.06, Bytes(1)),
    If(lambda ctx: find_save_version(ctx) >= 20.16, Bytes(8)),
    If(lambda ctx: find_save_version(ctx) >= 25.06, Bytes(21)),
    If(lambda ctx: find_save_version(ctx) >= 25.22, Bytes(4)),
    If(lambda ctx: find_save_version(ctx) >= 26.16, Bytes(8)),
    If(lambda ctx: find_save_version(ctx) >= 37, Bytes(3)),
    If(lambda ctx: find_save_version(ctx) >= 50, Bytes(8)),
    If(lambda ctx: find_save_version(ctx) >= 61.5, Flag),
    If(lambda ctx: find_save_version(ctx) >= 63, Bytes(5)),
    "unknown_count"/If(lambda ctx: find_save_version(ctx) >= 66.3, Int32ul),
    If(lambda ctx: find_save_version(ctx) >= 66.3, Bytes(12)),
    If(lambda ctx: find_save_version(ctx) >= 66.3, Array(lambda ctx: ctx.unknown_count, Bytes(4))),
    de_string,
    Bytes(5),
    If(lambda ctx: find_save_version(ctx) >= 13.13, Byte),
    If(lambda ctx: find_save_version(ctx) < 13.17, Struct(
        de_string,
        Int32ul,
        Bytes(4), # usually zero
    )),
    If(lambda ctx: find_save_version(ctx) >= 13.17, Bytes(2)),
    "ver37"/If(lambda ctx: find_save_version(ctx) >= 37, Struct(
        Int32ul,
        Int32ul
    ))
)
